########################################################################
# File name: test___init__.py
# This file is part of: aioxmpp
#
# LICENSE
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this program.  If not, see
# <http://www.gnu.org/licenses/>.
#
########################################################################
import unittest

import aioxmpp
import aioxmpp.disco as disco
import aioxmpp.disco.service as disco_service
import aioxmpp.disco.xso as disco_xso


class TestExports(unittest.TestCase):
    def test_DiscoClient(self):
        self.assertIs(disco.DiscoClient, disco_service.DiscoClient)
        self.assertIs(aioxmpp.DiscoClient, disco_service.DiscoClient)

    def test_DiscoServer(self):
        self.assertIs(disco.DiscoServer, disco_service.DiscoServer)
        self.assertIs(aioxmpp.DiscoServer, disco_service.DiscoServer)

    def test_xso(self):
        self.assertIs(disco.xso, disco_xso)

    def test_Node(self):
        self.assertIs(disco.Node, disco_service.Node)

    def test_StaticNode(self):
        self.assertIs(disco.StaticNode, disco_service.StaticNode)



########################################################################
# File name: test_service.py
# This file is part of: aioxmpp
#
# LICENSE
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this program.  If not, see
# <http://www.gnu.org/licenses/>.
#
########################################################################
import unittest
import unittest.mock

import aioxmpp.disco
import aioxmpp.mdr.service as mdr_service
import aioxmpp.mdr.xso as mdr_xso
import aioxmpp.service
import aioxmpp.tracking

from aioxmpp.testutils import (
    make_connected_client,
)


TEST_FROM = aioxmpp.JID.fromstr("juliet@capulet.example/balcony")
TEST_TO = aioxmpp.JID.fromstr("romeo@montague.example")


class TestDeliveryReceiptsService(unittest.TestCase):
    def setUp(self):
        self.cc = make_connected_client()
        self.disco_svr = aioxmpp.DiscoServer(self.cc)
        self.t = aioxmpp.tracking.MessageTracker()
        self.s = mdr_service.DeliveryReceiptsService(self.cc, dependencies={
            aioxmpp.DiscoServer: self.disco_svr,
        })
        self.msg = unittest.mock.Mock(spec=aioxmpp.Message)
        self.msg.xep0184_received = None
        self.msg.xep0184_request_receipt = False
        self.msg.to = TEST_TO
        self.msg.id_ = "foo"

    def tearDown(self):
        del self.s
        del self.cc

    def test_is_service(self):
        self.assertIsInstance(
            self.s,
            aioxmpp.service.Service,
        )

    def test_registers_inbound_message_filter(self):
        self.assertTrue(
            aioxmpp.service.is_inbound_message_filter(
                mdr_service.DeliveryReceiptsService._inbound_message_filter,
            )
        )

    def test_inbound_message_filter_returns_random_stanza(self):
        stanza = unittest.mock.Mock(spec=aioxmpp.Message)
        stanza.xep0184_received = None
        self.assertIs(
            self.s._inbound_message_filter(stanza),
            stanza,
        )

    def test_declares_disco_feature(self):
        self.assertIsInstance(
            mdr_service.DeliveryReceiptsService.disco_feature,
            aioxmpp.disco.register_feature,
        )
        self.assertEqual(
            mdr_service.DeliveryReceiptsService.disco_feature.feature,
            "urn:xmpp:receipts",
        )

    def test_attach_tracker_sets_xep0184_request(self):
        self.s.attach_tracker(self.msg, self.t)
        self.assertTrue(
            self.msg.xep0184_request_receipt,
        )

    def test_attach_tracker_autosets_id(self):
        self.s.attach_tracker(self.msg, self.t)
        self.msg.autoset_id.assert_called_once_with()

    def test_attach_tracker_returns_passed_tracker(self):
        t = self.s.attach_tracker(self.msg, self.t)
        self.assertIs(t, self.t)

    def test_attach_tracker_rejects_receipts(self):
        self.msg.xep0184_received = mdr_xso.Received("foo")
        with self.assertRaisesRegex(
                ValueError,
                r"requesting delivery receipts for delivery receipts is not "
                r"allowed"):
            self.s.attach_tracker(self.msg, self.t)

    def test_attach_tracker_rejects_errors(self):
        self.msg.type_ = aioxmpp.MessageType.ERROR
        with self.assertRaisesRegex(
                ValueError,
                r"requesting delivery receipts for errors is not supported"):
            self.s.attach_tracker(self.msg, self.t)

    def test_attach_tracker_creates_new_tracker_if_none_passed(self):
        with unittest.mock.patch(
                "aioxmpp.tracking.MessageTracker") as MessageTracker:
            t = self.s.attach_tracker(self.msg)

        MessageTracker.assert_called_once_with()
        self.assertEqual(t, MessageTracker())

    def test_set_tracker_state_to_DTR_on_ack_for_full_match(self):
        self.msg.to = self.msg.to.replace(resource="foo")
        self.s.attach_tracker(self.msg, self.t)

        ack = aioxmpp.Message(
            type_=aioxmpp.MessageType.CHAT,
            from_=TEST_TO.replace(resource="foo")
        )
        ack.xep0184_received = mdr_xso.Received(self.msg.id_)

        self.assertEqual(
            self.t.state,
            aioxmpp.tracking.MessageState.IN_TRANSIT,
        )
        self.assertIsNone(
            self.s._inbound_message_filter(ack)
        )

        self.assertEqual(
            self.t.state,
            aioxmpp.tracking.MessageState.DELIVERED_TO_RECIPIENT,
        )

    def test_handle_state_exception_from_tracker(self):
        self.msg.to = self.msg.to.replace(resource="foo")
        self.s.attach_tracker(self.msg, self.t)

        ack = aioxmpp.Message(
            type_=aioxmpp.MessageType.CHAT,
            from_=TEST_TO.replace(resource="foo")
        )
        ack.xep0184_received = mdr_xso.Received(self.msg.id_)

        self.t._set_state(aioxmpp.tracking.MessageState.SEEN_BY_RECIPIENT)

        self.assertIsNone(
            self.s._inbound_message_filter(ack)
        )

        self.assertEqual(
            self.t.state,
            aioxmpp.tracking.MessageState.SEEN_BY_RECIPIENT,
        )

    def test_do_not_modify_tracker_state_on_id_mismatch(self):
        self.msg.to = self.msg.to.replace(resource="foo")
        self.s.attach_tracker(self.msg, self.t)

        ack = aioxmpp.Message(
            type_=aioxmpp.MessageType.CHAT,
            from_=TEST_TO.replace(resource="foo")
        )
        ack.xep0184_received = mdr_xso.Received("fnord")

        self.assertEqual(
            self.t.state,
            aioxmpp.tracking.MessageState.IN_TRANSIT,
        )
        self.assertIsNone(
            self.s._inbound_message_filter(ack)
        )

        self.assertEqual(
            self.t.state,
            aioxmpp.tracking.MessageState.IN_TRANSIT,
        )

    def test_do_not_modify_tracker_state_on_bare_jid_mismatch(self):
        self.msg.to = self.msg.to.replace(resource="foo")
        self.s.attach_tracker(self.msg, self.t)

        ack = aioxmpp.Message(
            type_=aioxmpp.MessageType.CHAT,
            from_=self.msg.to.replace(localpart="nottheromeo")
        )
        ack.xep0184_received = mdr_xso.Received(self.msg.id_)

        self.assertEqual(
            self.t.state,
            aioxmpp.tracking.MessageState.IN_TRANSIT,
        )
        self.assertIsNone(
            self.s._inbound_message_filter(ack)
        )

        self.assertEqual(
            self.t.state,
            aioxmpp.tracking.MessageState.IN_TRANSIT,
        )


class Testcompose_receipt(unittest.TestCase):
    def setUp(self):
        self.msg = aioxmpp.Message(
            type_=aioxmpp.MessageType.CHAT,
            from_=TEST_FROM,
            to=TEST_TO,
        )
        self.msg.body[None] = "foo"
        self.msg.xep0184_request_receipt = True
        self.msg.autoset_id()

    def test_sends_to_bare_jid(self):
        msg = mdr_service.compose_receipt(self.msg)
        self.assertEqual(
            msg.to,
            self.msg.from_.bare()
        )

    def test_preserves_type(self):
        msg = mdr_service.compose_receipt(self.msg)
        self.assertEqual(msg.type_, self.msg.type_)

    def test_raises_on_error(self):
        with self.assertRaisesRegex(
                ValueError,
                r"receipts cannot be generated for error messages"):
            mdr_service.compose_receipt(
                aioxmpp.Message(
                    type_=aioxmpp.MessageType.ERROR
                )
            )

    def test_raises_on_receipt(self):
        self.msg.xep0184_received = mdr_xso.Received("foo")

        with self.assertRaisesRegex(
                ValueError,
                r"receipts cannot be generated for receipts"):
            mdr_service.compose_receipt(self.msg)

    def test_if_id_is_unset(self):
        self.msg.id_ = None

        with self.assertRaisesRegex(
                ValueError,
                r"receipts cannot be generated for id-less messages"):
            mdr_service.compose_receipt(self.msg)

    def test_receipt_refers_to_message_id(self):
        msg = mdr_service.compose_receipt(self.msg)

        self.assertIsInstance(
            msg.xep0184_received,
            mdr_xso.Received,
        )

        self.assertEqual(
            msg.xep0184_received.message_id,
            self.msg.id_,
        )

    def test_strips_body(self):
        msg = mdr_service.compose_receipt(self.msg)
        self.assertFalse(msg.body)



#!/usr/bin/env python3
########################################################################
# File name: xmpp_bridge.py
# This file is part of: aioxmpp
#
# LICENSE
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this program.  If not, see
# <http://www.gnu.org/licenses/>.
#
########################################################################
import asyncio
import asyncio.streams
import os
import signal
import sys

import aioxmpp


async def stdout_writer():
    """
    This is a bit complex, as stdout can be a pipe or a file.

    If it is a file, we cannot use
    :meth:`asycnio.BaseEventLoop.connect_write_pipe`.
    """
    if sys.stdout.seekable():
        # it’s a file
        return sys.stdout.buffer.raw

    if os.isatty(sys.stdin.fileno()):
        # it’s a tty, use fd 0
        fd_to_use = 0
    else:
        fd_to_use = 1

    twrite, pwrite = await loop.connect_write_pipe(
        asyncio.streams.FlowControlMixin,
        os.fdopen(fd_to_use, "wb"),
    )

    swrite = asyncio.StreamWriter(
        twrite,
        pwrite,
        None,
        loop,
    )

    return swrite


async def main(local, password, peer,
               strip_newlines, add_newlines):
    loop = asyncio.get_event_loop()
    swrite = await stdout_writer()

    sread = asyncio.StreamReader()
    tread, pread = await loop.connect_read_pipe(
        lambda: asyncio.StreamReaderProtocol(sread),
        sys.stdin,
    )

    client = aioxmpp.Client(
        local,
        aioxmpp.make_security_layer(
            password,
        )
    )

    sigint = asyncio.Event()
    loop.add_signal_handler(signal.SIGINT, sigint.set)
    loop.add_signal_handler(signal.SIGTERM, sigint.set)

    def recv(message):
        body = message.body.lookup(
            [aioxmpp.structs.LanguageRange.fromstr("*")]
        )
        if add_newlines:
            body += "\n"
        swrite.write(body.encode("utf-8"))

    client.stream.register_message_callback(
        "chat",
        peer,
        recv
    )

    sigint_future = asyncio.ensure_future(sigint.wait())
    read_future = asyncio.ensure_future(sread.readline())

    try:
        async with client.connected() as stream:
            # we send directed presence to the peer
            pres = aioxmpp.Presence(
                type_=aioxmpp.PresenceType.AVAILABLE,
                to=peer,
            )
            await stream.send(pres)
            while True:
                done, pending = await asyncio.wait(
                    [
                        sigint_future,
                        read_future,
                    ],
                    return_when=asyncio.FIRST_COMPLETED,
                )

                if sigint_future in done:
                    break

                if read_future in done:
                    line = read_future.result().decode()
                    if not line:
                        break

                    if strip_newlines:
                        line = line.rstrip()

                    msg = aioxmpp.Message(
                        type_="chat",
                        to=peer,
                    )
                    msg.body[None] = line

                    await stream.send(msg)

                    read_future = asyncio.ensure_future(
                        sread.readline()
                    )

    finally:
        sigint_future.cancel()
        read_future.cancel()


def jid(s):
    return aioxmpp.JID.fromstr(s)


if __name__ == "__main__":
    import argparse

    parser = argparse.ArgumentParser(
        description="""
        Send lines from stdin to the given peer and print messages received
        from the peer to stdout.""",
        epilog="""
        The password must be set in the XMPP_BRIDGE_PASSWORD environment
        variable."""
    )

    parser.add_argument(
        "--no-strip-newlines",
        dest="strip_newlines",
        action="store_false",
        default=True,
        help="Disable stripping newlines from stdin"
    )

    parser.add_argument(
        "--no-add-newlines",
        dest="add_newlines",
        action="store_false",
        default=True,
        help="Disable adding newlines to stdout"
    )

    parser.add_argument(
        "local",
        help="JID to bind to",
        type=jid,
    )
    parser.add_argument(
        "peer",
        help="JID of the peer to send messages to",
        type=jid,
    )

    args = parser.parse_args()

    try:
        password = os.environ["XMPP_BRIDGE_PASSWORD"]
    except KeyError:
        parser.print_help()
        print("XMPP_BRIDGE_PASSWORD is not set", file=sys.stderr)
        sys.exit(1)

    loop = asyncio.get_event_loop()
    loop.run_until_complete(main(
        args.local, password, args.peer,
        args.strip_newlines,
        args.add_newlines,
    ))
    loop.close()



########################################################################
# File name: __init__.py
# This file is part of: aioxmpp
#
# LICENSE
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this program.  If not, see
# <http://www.gnu.org/licenses/>.
#
########################################################################
"""
:mod:`~aioxmpp.httpupload` --- HTTP Upload support (:xep:`363`)
###############################################################

The :xep:`363` HTTP Upload protocol allows an XMPP client to obtain a PUT and
GET URL for storage on the server. It can upload a file (once) using the PUT
URL and distribute access to the file via the GET url.

This module does *not* handle the HTTP part of the interaction. We recommend
to use :mod:`aiohttp` for this, but you can use any HTTP library which supports
GET, PUT and sending custom headers.

Example use::

    client = <your aioxmpp.Client>
    slot = await client.send(aioxmpp.httpupload.Request(
        filename,
        size,
        content_type,
    ))
    # http_put_file is provided by you via an HTTP library
    await http_put_file(
        slot.put.url,
        slot.put.headers,
        filename
    )

.. autofunction:: request_slot

.. autoclass:: Request

.. module:: aioxmpp.httpupload.xso

.. currentmodule:: aioxmpp.httpupload.xso

.. autoclass:: Slot()

.. autoclass:: Get()

.. autoclass:: Put()
"""
import asyncio

from ..structs import JID, IQType
from ..stanza import IQ
from .xso import Request


@asyncio.coroutine
def request_slot(client,
                 service: JID,
                 filename: str,
                 size: int,
                 content_type: str):
    """
    Request an HTTP upload slot.

    :param client: The client to request the slot with.
    :type client: :class:`aioxmpp.Client`
    :param service: Address of the HTTP upload service.
    :type service: :class:`~aioxmpp.JID`
    :param filename: Name of the file (without path), may be used by the server
        to generate the URL.
    :type filename: :class:`str`
    :param size: Size of the file in bytes
    :type size: :class:`int`
    :param content_type: The MIME type of the file
    :type content_type: :class:`str`
    :return: The assigned upload slot.
    :rtype: :class:`.xso.Slot`

    Sends a :xep:`363` slot request to the XMPP service to obtain HTTP
    PUT and GET URLs for a file upload.

    The upload slot is returned as a :class:`~.xso.Slot` object.
    """

    payload = Request(filename, size, content_type)
    return (yield from client.send(IQ(
        type_=IQType.GET,
        to=service,
        payload=payload
    )))



########################################################################
# File name: service.py
# This file is part of: aioxmpp
#
# LICENSE
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this program.  If not, see
# <http://www.gnu.org/licenses/>.
#
########################################################################
import asyncio

import aioxmpp.disco
import aioxmpp.service
import aioxmpp.structs

from aioxmpp.utils import namespaces

from . import xso as ping_xso


class PingService(aioxmpp.service.Service):
    """
    Service implementing XMPP Ping (:xep:`199`).

    This service implements the response to XMPP Pings and provides a method
    to send pings.

    .. automethod:: ping
    """

    ORDER_AFTER = [
        aioxmpp.disco.DiscoServer,
    ]

    _ping_feature = aioxmpp.disco.register_feature(
        namespaces.xep0199_ping
    )

    @aioxmpp.service.iq_handler(aioxmpp.structs.IQType.GET, ping_xso.Ping)
    @asyncio.coroutine
    def handle_ping(self, request):
        return ping_xso.Ping()

    @asyncio.coroutine
    def ping(self, peer):
        """
        Ping a peer.

        :param peer: The peer to ping.
        :type peer: :class:`aioxmpp.JID`
        :raises aioxmpp.errors.XMPPError: as received

        Send a :xep:`199` ping IQ to `peer` and wait for the reply.

        .. note::

            If the peer does not support :xep:`199`, they will respond with
            a ``cancel`` ``service-unavailable`` error. However, some
            implementations return a ``cancel`` ``feature-not-implemented``
            error instead. Callers should be prepared for the
            :class:`aioxmpp.XMPPCancelError` exceptions in those cases.
        """

        iq = aioxmpp.IQ(
            to=peer,
            type_=aioxmpp.IQType.GET,
            payload=ping_xso.Ping()
        )

        yield from self.client.send(iq)



########################################################################
# File name: pars.py
# This file is part of: aioxmpp
#
# LICENSE
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this program.  If not, see
# <http://www.gnu.org/licenses/>.
#
########################################################################
import aioxmpp.xso as xso

from aioxmpp.utils import namespaces

from ..stanza import Presence


namespaces.xep0379_pars = "urn:xmpp:pars:0"


class Preauth(xso.XSO):
    """
    The preauth element for :xep:`Pre-Authenticated Roster Subcription <379>`.

    .. attribute:: token

       The pre-auth token associated with this subscription request.
    """
    TAG = namespaces.xep0379_pars, "preauth"

    token = xso.Attr(
        "token",
        type_=xso.String(),
    )


Presence.xep0379_preauth = xso.Child([Preauth])



########################################################################
# File name: test_query.py
# This file is part of: aioxmpp
#
# LICENSE
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this program.  If not, see
# <http://www.gnu.org/licenses/>.
#
########################################################################
import contextlib
import copy
import itertools
import operator
import unittest
import unittest.mock

import aioxmpp.xso.query as xso_query
import aioxmpp.xso as xso


class FooXSO(xso.XSO):
    TAG = (None, "foo")

    attr = xso.Attr(
        "attr"
    )


class BarXSO(xso.XSO):
    TAG = (None, "bar")

    child = xso.Child([
        FooXSO,
    ])


class BazXSO(FooXSO):
    TAG = (None, "baz")

    attr2 = xso.Attr(
        "attr2"
    )


class RootXSO(xso.XSO):
    TAG = (None, "root")

    children = xso.ChildList([
        FooXSO,
        BarXSO,
    ])

    attr = xso.Attr(
        "attr"
    )


class TestEvaluationContext(unittest.TestCase):
    def setUp(self):
        self.ec = xso_query.EvaluationContext()

    def tearDown(self):
        del self.ec

    def test_get_toplevel_object_raises_KeyError_for_undefined(self):
        class Foo:
            pass

        with self.assertRaises(KeyError):
            self.ec.get_toplevel_object(Foo)

    def test_get_toplevel_object_returns_set_object(self):
        class Foo:
            pass

        self.ec.set_toplevel_object(
            unittest.mock.sentinel.instance,
            class_=Foo,
        )

        self.assertIs(
            self.ec.get_toplevel_object(Foo),
            unittest.mock.sentinel.instance
        )

    def test_set_toplevel_object_autodetects_type(self):
        class Foo:
            pass

        instance = Foo()

        self.ec.set_toplevel_object(
            instance,
        )

        self.assertIs(
            self.ec.get_toplevel_object(Foo),
            instance
        )

        with self.assertRaises(KeyError):
            self.ec.get_toplevel_object(object())

    def test_copy_copies_toplevels(self):
        obj1 = object()

        self.ec.set_toplevel_object(obj1)
        ec2 = copy.copy(self.ec)
        self.assertIs(
            self.ec.get_toplevel_object(object),
            ec2.get_toplevel_object(object),
        )

        ec2.set_toplevel_object(object())

        self.assertIsNot(
            self.ec.get_toplevel_object(object),
            ec2.get_toplevel_object(object),
        )
        self.assertIs(
            self.ec.get_toplevel_object(object),
            obj1
        )

    def test_eval_simply_forwards_result(self):
        expr = unittest.mock.Mock()

        result = self.ec.eval(expr)

        expr.eval.assert_called_with(self.ec)

        self.assertEqual(result, expr.eval())

    def test_eval_bool_with_false_generator(self):
        def generator():
            return
            yield

        expr = unittest.mock.Mock()
        expr.eval.return_value = generator()

        result = self.ec.eval_bool(expr)

        expr.eval.assert_called_with(self.ec)

        self.assertFalse(result)

    def test_eval_bool_with_true_generator(self):
        aborted = False

        def generator():
            nonlocal aborted
            try:
                yield
            except GeneratorExit:
                aborted = True
                raise

        expr = unittest.mock.Mock()
        expr.eval.return_value = generator()

        result = self.ec.eval_bool(expr)

        expr.eval.assert_called_with(self.ec)

        self.assertTrue(result)
        self.assertTrue(aborted)

    def test_eval_bool_with_false_sequence(self):
        expr = unittest.mock.Mock()
        expr.eval.return_value = []

        result = self.ec.eval_bool(expr)

        expr.eval.assert_called_with(self.ec)

        self.assertFalse(result)

    def test_eval_bool_with_true_sequence(self):
        expr = unittest.mock.Mock()
        expr.eval.return_value = ["1"]

        result = self.ec.eval_bool(expr)

        expr.eval.assert_called_with(self.ec)

        self.assertTrue(result)


class Test_SoftExprMixin(unittest.TestCase):
    def test_getitem_uses_as_expr(self):
        index = unittest.mock.sentinel.index
        e = xso_query._SoftExprMixin()

        with contextlib.ExitStack() as stack:
            as_expr = stack.enter_context(
                unittest.mock.patch("aioxmpp.xso.query.as_expr")
            )

            Nth = stack.enter_context(
                unittest.mock.patch("aioxmpp.xso.query.Nth")
            )

            result = e[index]

        as_expr.assert_called_with(index)
        Nth.assert_called_with(
            e,
            as_expr(),
        )

        self.assertEqual(
            result,
            Nth(),
        )

    def test_getitem_with_number(self):
        e = xso_query._SoftExprMixin()

        with contextlib.ExitStack() as stack:
            Nth = stack.enter_context(
                unittest.mock.patch(
                    "aioxmpp.xso.query.Nth",
                )
            )

            Constant = stack.enter_context(
                unittest.mock.patch(
                    "aioxmpp.xso.query.Constant",
                )
            )

            result = e[10]

        Constant.assert_called_with(
            10,
        )

        Nth.assert_called_with(
            e,
            Constant()
        )

        self.assertEqual(
            result,
            Nth()
        )

    def test_getitem_with_slice(self):
        e = xso_query._SoftExprMixin()

        with contextlib.ExitStack() as stack:
            Nth = stack.enter_context(
                unittest.mock.patch(
                    "aioxmpp.xso.query.Nth",
                )
            )

            Constant = stack.enter_context(
                unittest.mock.patch(
                    "aioxmpp.xso.query.Constant",
                )
            )

            result = e[1:2:3]

        Constant.assert_called_with(
            slice(1, 2, 3),
        )

        Nth.assert_called_with(
            e,
            Constant()
        )

        self.assertEqual(
            result,
            Nth()
        )

    def test_expr_slash_PreExpr(self):
        e = xso_query._SoftExprMixin()

        pe = unittest.mock.Mock(spec=xso_query.PreExpr)

        with unittest.mock.patch(
                "aioxmpp.xso.query.as_expr") as as_expr:
            result = e / pe

        as_expr.assert_called_with(pe, lhs=e)
        self.assertEqual(
            result,
            as_expr(),
        )

    def test_expr_slash_Expr(self):
        e = xso_query._SoftExprMixin()

        e2 = unittest.mock.Mock(spec=xso_query.Expr)

        with unittest.mock.patch(
                "aioxmpp.xso.query.as_expr") as as_expr:
            result = e / e2

        as_expr.assert_called_with(e2, lhs=e)
        self.assertEqual(
            result,
            as_expr(),
        )

    def test_expr_getitem_where(self):
        expr = xso_query._SoftExprMixin()

        w = unittest.mock.Mock(spec=xso_query.where)
        w.expr = unittest.mock.Mock(spec=xso_query.Expr)

        with unittest.mock.patch(
                "aioxmpp.xso.query.ExprFilter") as ExprFilter:
            result = expr[w]

        ExprFilter.assert_called_with(expr, w.expr)
        self.assertEqual(
            result,
            ExprFilter(),
        )


class Test_ExprMixin(unittest.TestCase):
    def _test_cmp_op(self, operator_):
        expr1 = xso_query._ExprMixin()
        expr2 = xso_query._ExprMixin()

        def as_exprs():
            yield unittest.mock.sentinel.expr1
            yield unittest.mock.sentinel.expr2

        with contextlib.ExitStack() as stack:
            as_expr = stack.enter_context(
                unittest.mock.patch("aioxmpp.xso.query.as_expr")
            )
            as_expr.side_effect = as_exprs()

            CmpOp = stack.enter_context(
                unittest.mock.patch("aioxmpp.xso.query.CmpOp")
            )

            result = operator_(expr1, expr2)

        self.assertSequenceEqual(
            as_expr.mock_calls,
            [
                unittest.mock.call(expr1),
                unittest.mock.call(expr2),
            ]
        )

        CmpOp.assert_called_with(
            unittest.mock.sentinel.expr1,
            unittest.mock.sentinel.expr2,
            operator_,
        )

        self.assertEqual(
            result,
            CmpOp(),
        )

    def test_eq(self):
        self._test_cmp_op(operator.eq)

    def test_ne(self):
        self._test_cmp_op(operator.ne)

    def test_lt(self):
        self._test_cmp_op(operator.lt)

    def test_le(self):
        self._test_cmp_op(operator.le)

    def test_gt(self):
        self._test_cmp_op(operator.gt)

    def test_ge(self):
        self._test_cmp_op(operator.ge)

    def test_not(self):
        expr1 = xso_query._ExprMixin()

        def as_exprs():
            yield unittest.mock.sentinel.expr1

        with contextlib.ExitStack() as stack:
            as_expr = stack.enter_context(
                unittest.mock.patch("aioxmpp.xso.query.as_expr")
            )
            as_expr.side_effect = as_exprs()

            NotOp = stack.enter_context(
                unittest.mock.patch("aioxmpp.xso.query.NotOp")
            )

            result = xso_query.not_(expr1)

        self.assertSequenceEqual(
            as_expr.mock_calls,
            [
                unittest.mock.call(expr1),
            ]
        )

        NotOp.assert_called_with(
            unittest.mock.sentinel.expr1,
        )

        self.assertEqual(
            result,
            NotOp(),
        )


class TestExpr(unittest.TestCase):
    def setUp(self):
        class DummyExpr(xso_query.Expr):
            def __init__(self, *args, mock=None, **kwargs):
                super().__init__(*args, **kwargs)
                self.__mock = mock

            def eval(self, ec):
                return self.__mock.eval(ec)

        self.DummyExpr = DummyExpr

    def tearDown(self):
        del self.DummyExpr

    def test_is_abstract(self):
        with self.assertRaisesRegex(TypeError, "abstract"):
            xso_query.Expr()

    def test_has_ExprMixin(self):
        self.assertTrue(issubclass(
            xso_query.Expr,
            xso_query._ExprMixin,
        ))

    def test_eval_leaf_uses_eval(self):
        expr = unittest.mock.Mock()
        ec = unittest.mock.sentinel.ec

        e = self.DummyExpr(mock=expr)
        result = e.eval_leaf(ec)

        expr.eval.assert_called_with(ec)

        self.assertEqual(result, expr.eval())

    def test_eval_leaf_evaluates_generator(self):
        def generator():
            yield 1
            yield 2

        expr = unittest.mock.Mock()
        expr.eval.return_value = generator()
        ec = unittest.mock.sentinel.ec

        e = self.DummyExpr(mock=expr)
        result = e.eval_leaf(ec)

        expr.eval.assert_called_with(ec)

        self.assertSequenceEqual(
            result,
            [1, 2]
        )

    def test_eval_leaf_keeps_mappings(self):
        expr = unittest.mock.Mock()
        expr.eval.return_value = {"1": "2"}
        ec = unittest.mock.sentinel.ec

        e = self.DummyExpr(mock=expr)
        result = e.eval_leaf(ec)

        expr.eval.assert_called_with(ec)

        self.assertSequenceEqual(
            result,
            {"1": "2"}
        )


class TestContextInstance(unittest.TestCase):
    def setUp(self):
        class Cls:
            pass

        self.Cls = Cls
        self.toplevel = Cls()

        self.ec = unittest.mock.Mock()
        self.ec.get_toplevel_object.return_value = self.toplevel

    def tearDown(self):
        del self.ec
        del self.Cls
        del self.toplevel

    def test_is_Expr(self):
        self.assertTrue(issubclass(
            xso_query.ContextInstance,
            xso_query.Expr,
        ))

    def test_eval_returns_toplevel_object_of_class(self):
        ci = xso_query.ContextInstance(self.Cls)

        self.assertEqual(
            ci.eval(self.ec),
            [self.toplevel],
        )

        self.ec.get_toplevel_object.assert_called_with(self.Cls)

    def test_eval_returns_empty_sequence_for_unset_toplevel(self):
        self.ec.get_toplevel_object.side_effect = KeyError

        ci = xso_query.ContextInstance(self.Cls)

        self.assertEqual(
            ci.eval(self.ec),
            [],
        )

        self.ec.get_toplevel_object.assert_called_with(self.Cls)


class TestGetDescriptor(unittest.TestCase):
    def setUp(self):
        pass

    def tearDown(self):
        pass

    def test_is_Expr(self):
        self.assertTrue(issubclass(
            xso_query.GetDescriptor,
            xso_query.Expr,
        ))

    def test_update_values_appends(self):
        v = unittest.mock.Mock()
        vnew = unittest.mock.Mock()

        gd = xso_query.GetDescriptor(
            unittest.mock.sentinel.expr,
            unittest.mock.sentinel.descriptor,
        )

        gd.update_values(v, vnew)

        v.append.assert_called_with(vnew)

    def test_new_values_creates_empty_sequence(self):
        gd = xso_query.GetDescriptor(
            unittest.mock.sentinel.expr,
            unittest.mock.sentinel.descriptor,
        )

        s1 = gd.new_values()
        s2 = gd.new_values()

        self.assertSequenceEqual(
            s1,
            []
        )

        self.assertSequenceEqual(
            s2,
            []
        )

        self.assertIsNot(s1, s2)

    def test_eval_gets_descriptor(self):
        class Bar:
            pass

        i1 = Bar()
        i2 = Bar()

        def values():
            yield unittest.mock.sentinel.v1
            yield unittest.mock.sentinel.v2

        descriptor = unittest.mock.PropertyMock()
        descriptor.side_effect = values()

        expr = unittest.mock.Mock()
        ec = unittest.mock.sentinel

        expr.eval.return_value = [
            i1,
            i2,
        ]

        gd = xso_query.GetDescriptor(expr, descriptor)

        with contextlib.ExitStack() as stack:
            new_values = stack.enter_context(
                unittest.mock.patch.object(gd, "new_values")
            )

            update_values = stack.enter_context(
                unittest.mock.patch.object(gd, "update_values")
            )

            new_values.return_value = unittest.mock.sentinel.vs

            result = gd.eval(ec)

        self.assertEqual(
            result,
            unittest.mock.sentinel.vs,
        )

        self.assertSequenceEqual(
            update_values.mock_calls,
            [
                unittest.mock.call(
                    unittest.mock.sentinel.vs,
                    unittest.mock.sentinel.v1,
                ),
                unittest.mock.call(
                    unittest.mock.sentinel.vs,
                    unittest.mock.sentinel.v2,
                ),
            ]
        )

        expr.eval.assert_called_with(ec)

    def test_eval_skips_on_AttributeError(self):
        class Bar:
            pass

        i1 = Bar()
        i2 = Bar()

        def values():
            yield AttributeError
            yield unittest.mock.sentinel.v2

        descriptor = unittest.mock.PropertyMock()
        descriptor.side_effect = values()

        expr = unittest.mock.Mock()
        ec = unittest.mock.sentinel

        expr.eval.return_value = [
            i1,
            i2,
        ]

        gd = xso_query.GetDescriptor(expr, descriptor)

        with contextlib.ExitStack() as stack:
            new_values = stack.enter_context(
                unittest.mock.patch.object(gd, "new_values")
            )

            update_values = stack.enter_context(
                unittest.mock.patch.object(gd, "update_values")
            )

            new_values.return_value = unittest.mock.sentinel.vs

            result = gd.eval(ec)

        self.assertEqual(
            result,
            unittest.mock.sentinel.vs,
        )

        self.assertSequenceEqual(
            update_values.mock_calls,
            [
                unittest.mock.call(
                    unittest.mock.sentinel.vs,
                    unittest.mock.sentinel.v2,
                ),
            ]
        )

        expr.eval.assert_called_with(ec)


class TestGetMappingDescriptor(unittest.TestCase):
    def test_is_GetDescriptor(self):
        self.assertTrue(issubclass(
            xso_query.GetMappingDescriptor,
            xso_query.GetDescriptor,
        ))

    def test_new_values_returns_result_of_mapping_factory(self):
        expr = unittest.mock.sentinel.expr
        descriptor = unittest.mock.sentinel.descriptor
        mf = unittest.mock.Mock()

        gmd = xso_query.GetMappingDescriptor(
            expr,
            descriptor,
            mapping_factory=mf
        )

        vs = gmd.new_values()
        mf.assert_called_with()
        self.assertEqual(vs, mf())

    def test_update_values_uses_update(self):
        expr = unittest.mock.sentinel.expr
        descriptor = unittest.mock.sentinel.descriptor

        v = unittest.mock.Mock()
        vnew = unittest.mock.sentinel.vnew

        gmd = xso_query.GetMappingDescriptor(
            expr,
            descriptor,
        )

        gmd.update_values(v, vnew)

        v.update.assert_called_with(vnew)


class TestGetSequenceDescriptor(unittest.TestCase):
    def test_is_GetDescriptor(self):
        self.assertTrue(issubclass(
            xso_query.GetSequenceDescriptor,
            xso_query.GetDescriptor,
        ))

    def test_new_values_returns_result_of_sequence_factory(self):
        expr = unittest.mock.sentinel.expr
        descriptor = unittest.mock.sentinel.descriptor
        sf = unittest.mock.Mock()

        gsd = xso_query.GetSequenceDescriptor(
            expr,
            descriptor,
            sequence_factory=sf
        )

        vs = gsd.new_values()
        sf.assert_called_with()
        self.assertEqual(vs, sf())

    def test_update_values_uses_extend(self):
        expr = unittest.mock.sentinel.expr
        descriptor = unittest.mock.sentinel.descriptor

        v = unittest.mock.Mock()
        vnew = unittest.mock.sentinel.vnew

        gsd = xso_query.GetSequenceDescriptor(
            expr,
            descriptor,
        )

        gsd.update_values(v, vnew)

        v.extend.assert_called_with(vnew)


class TestGetInstances(unittest.TestCase):
    def test_is_Expr(self):
        self.assertTrue(issubclass(
            xso_query.GetInstances,
            xso_query.Expr,
        ))

    def test_eval_filters_for_class(self):
        class Foo:
            pass

        class Bar:
            pass

        expr = unittest.mock.Mock()
        ec = unittest.mock.sentinel

        gi = xso_query.GetInstances(expr, Bar)

        expr.eval.return_value = [
            Foo(),
            Foo(),
            Foo(),
        ]

        self.assertSequenceEqual(
            list(gi.eval(ec)),
            [],
        )

        expr.eval.assert_called_with(ec)

    def test_eval_returns_instances(self):
        class Foo:
            pass

        class Bar:
            pass

        i1 = Bar()
        i2 = Bar()

        expr = unittest.mock.Mock()
        ec = unittest.mock.sentinel

        gi = xso_query.GetInstances(expr, Bar)

        expr.eval.return_value = [
            i1,
            Foo(),
            Foo(),
            Foo(),
            i2,
        ]

        self.assertSequenceEqual(
            list(gi.eval(ec)),
            [i1, i2],
        )

        expr.eval.assert_called_with(ec)


class TestNth(unittest.TestCase):
    def test_is_Expr(self):
        self.assertTrue(issubclass(
            xso_query.Nth,
            xso_query.Expr,
        ))

    def test_eval_returns_only_nth(self):
        expr = unittest.mock.Mock()
        nth_expr = unittest.mock.Mock()
        nth_expr.eval.return_value = [2]

        ec = unittest.mock.sentinel

        nth = xso_query.Nth(expr, nth_expr)

        expr.eval.return_value = [
            getattr(unittest.mock.sentinel, "v{}".format(i))
            for i in range(10)
        ]

        self.assertSequenceEqual(
            list(nth.eval(ec)),
            [unittest.mock.sentinel.v2]
        )

        nth_expr.eval.assert_called_with(ec)
        expr.eval.assert_called_with(ec)

    def test_eval_works_with_slice(self):
        expr = unittest.mock.Mock()
        nth_expr = unittest.mock.Mock()
        nth_expr.eval.return_value = [slice(1, 8, 2)]
        ec = unittest.mock.sentinel

        nth = xso_query.Nth(expr, nth_expr)

        expr.eval.return_value = [
            getattr(unittest.mock.sentinel, "v{}".format(i))
            for i in range(10)
        ]

        self.assertSequenceEqual(
            list(nth.eval(ec)),
            [
                unittest.mock.sentinel.v1,
                unittest.mock.sentinel.v3,
                unittest.mock.sentinel.v5,
                unittest.mock.sentinel.v7,
            ]
        )

        nth_expr.eval.assert_called_with(ec)
        expr.eval.assert_called_with(ec)


class TestExprFilter(unittest.TestCase):
    def test_is_Expr(self):
        self.assertTrue(issubclass(
            xso_query.ExprFilter,
            xso_query.Expr,
        ))

    def test_eval(self):
        ec = unittest.mock.Mock()
        filter_expr = unittest.mock.Mock()
        expr = unittest.mock.Mock()

        ef = xso_query.ExprFilter(
            expr,
            filter_expr
        )

        expr.eval.return_value = [
            unittest.mock.sentinel.v1,
            unittest.mock.sentinel.v2,
        ]

        def values():
            yield True
            yield False

        with unittest.mock.patch("copy.copy") as copy_:
            copy_().eval_bool.side_effect = values()
            copy_.mock_calls.clear()

            result = ef.eval(ec)

            self.assertSequenceEqual(
                list(result),
                [
                    unittest.mock.sentinel.v1,
                ]
            )

        expr.eval.assert_called_with(ec)

        self.assertSequenceEqual(
            copy_.mock_calls,
            [
                unittest.mock.call(ec),
                unittest.mock.call().set_toplevel_object(
                    unittest.mock.sentinel.v1,
                ),
                unittest.mock.call().eval_bool(
                    filter_expr,
                ),
                unittest.mock.call(ec),
                unittest.mock.call().set_toplevel_object(
                    unittest.mock.sentinel.v2,
                ),
                unittest.mock.call().eval_bool(
                    filter_expr,
                ),
            ]
        )


class Testwhere(unittest.TestCase):
    def test_init(self):
        w = xso_query.where(unittest.mock.sentinel.expr)
        self.assertIs(w.expr, unittest.mock.sentinel.expr)


class TestCmpOp(unittest.TestCase):
    def test_is_Expr(self):
        self.assertTrue(issubclass(
            xso_query.CmpOp,
            xso_query.Expr,
        ))

    def test_eval_leaf_returns_true_if_any_matches(self):
        ec = unittest.mock.sentinel.ec

        def results():
            yield False
            yield True
            yield False

        op = unittest.mock.Mock()
        op.side_effect = results()

        expr = unittest.mock.Mock()
        expr.eval_leaf.return_value = [
            unittest.mock.sentinel.v1,
            unittest.mock.sentinel.v2,
            unittest.mock.sentinel.v3,
        ]

        ref_expr = unittest.mock.Mock()
        ref_expr.eval_leaf.return_value = [
            unittest.mock.sentinel.rv
        ]

        co = xso_query.CmpOp(expr, ref_expr, op)

        result = co.eval_leaf(ec)

        self.assertTrue(result)

        self.assertSequenceEqual(
            expr.mock_calls,
            [
                unittest.mock.call.eval_leaf(ec),
            ]
        )

        self.assertSequenceEqual(
            ref_expr.mock_calls,
            [
                unittest.mock.call.eval_leaf(ec),
            ]
        )

        self.assertSequenceEqual(
            op.mock_calls,
            [
                unittest.mock.call(
                    unittest.mock.sentinel.v1,
                    unittest.mock.sentinel.rv,
                ),
                unittest.mock.call(
                    unittest.mock.sentinel.v2,
                    unittest.mock.sentinel.rv,
                ),
            ]
        )

    def test_eval_leaf_returns_false_if_none_matches(self):
        ec = unittest.mock.sentinel.ec

        def results():
            return
            yield

        op = unittest.mock.Mock()
        op.side_effect = results()

        expr = unittest.mock.Mock()
        expr.eval_leaf.return_value = [
        ]

        ref_expr = unittest.mock.Mock()
        ref_expr.eval_leaf.return_value = [
            unittest.mock.sentinel.rv
        ]

        co = xso_query.CmpOp(expr, ref_expr, op)

        result = co.eval_leaf(ec)

        self.assertFalse(result)

        self.assertSequenceEqual(
            expr.mock_calls,
            [
                unittest.mock.call.eval_leaf(ec),
            ]
        )

        self.assertSequenceEqual(
            ref_expr.mock_calls,
            [
                unittest.mock.call.eval_leaf(ec),
            ]
        )

        self.assertSequenceEqual(
            op.mock_calls,
            [
            ]
        )

    def test_eval_leaf_returns_false_if_no_ref_values(self):
        ec = unittest.mock.sentinel.ec

        def results():
            return
            yield

        op = unittest.mock.Mock()
        op.side_effect = results()

        expr = unittest.mock.Mock()
        expr.eval_leaf.return_value = [
            unittest.mock.sentinel.v1,
            unittest.mock.sentinel.v2,
            unittest.mock.sentinel.v3,
        ]

        ref_expr = unittest.mock.Mock()
        ref_expr.eval_leaf.return_value = [
        ]

        co = xso_query.CmpOp(expr, ref_expr, op)

        result = co.eval_leaf(ec)

        self.assertFalse(result)

        self.assertSequenceEqual(
            expr.mock_calls,
            [
                unittest.mock.call.eval_leaf(ec),
            ]
        )

        self.assertSequenceEqual(
            ref_expr.mock_calls,
            [
                unittest.mock.call.eval_leaf(ec),
            ]
        )

        self.assertSequenceEqual(
            op.mock_calls,
            [
            ]
        )

    def test_eval_leaf_cartesian_product(self):
        ec = unittest.mock.sentinel.ec

        def results():
            while True:
                yield False

        op = unittest.mock.Mock()
        op.side_effect = results()

        expr = unittest.mock.Mock()
        expr.eval_leaf.return_value = [
            unittest.mock.sentinel.v1,
            unittest.mock.sentinel.v2,
            unittest.mock.sentinel.v3,
        ]

        ref_expr = unittest.mock.Mock()
        ref_expr.eval_leaf.return_value = [
            unittest.mock.sentinel.rv1,
            unittest.mock.sentinel.rv2,
            unittest.mock.sentinel.rv3,
        ]

        co = xso_query.CmpOp(expr, ref_expr, op)

        result = co.eval_leaf(ec)

        self.assertFalse(result)

        self.assertSequenceEqual(
            expr.mock_calls,
            [
                unittest.mock.call.eval_leaf(ec),
            ]
        )

        self.assertSequenceEqual(
            ref_expr.mock_calls,
            [
                unittest.mock.call.eval_leaf(ec),
            ]
        )

        self.assertSequenceEqual(
            op.mock_calls,
            [
                unittest.mock.call(v1, v2)
                for v1, v2 in itertools.product(
                    expr.eval_leaf.return_value,
                    ref_expr.eval_leaf.return_value,
                )
            ]
        )

    def test_eval_returns_empty_list_on_false(self):
        ec = unittest.mock.sentinel.ec
        expr = unittest.mock.sentinel.expr
        ref_expr = unittest.mock.sentinel.ref_expr
        op = unittest.mock.sentinel.op

        co = xso_query.CmpOp(expr, ref_expr, op)

        with unittest.mock.patch.object(co, "eval_leaf") as eval_leaf:
            eval_leaf.return_value = False

            self.assertSequenceEqual(
                list(co.eval(ec)),
                []
            )

        eval_leaf.assert_called_with(ec)

    def test_eval_returns_list_with_single_True_on_true(self):
        ec = unittest.mock.sentinel.ec
        expr = unittest.mock.sentinel.expr
        ref_expr = unittest.mock.sentinel.ref_expr
        op = unittest.mock.sentinel.op

        co = xso_query.CmpOp(expr, ref_expr, op)

        with unittest.mock.patch.object(co, "eval_leaf") as eval_leaf:
            eval_leaf.return_value = True

            self.assertSequenceEqual(
                list(co.eval(ec)),
                [True]
            )

        eval_leaf.assert_called_with(ec)


class TestNotOp(unittest.TestCase):
    def test_is_Expr(self):
        self.assertTrue(issubclass(
            xso_query.NotOp,
            xso_query.Expr,
        ))

    def test_eval_leaf_uses_eval_bool(self):
        ec = unittest.mock.Mock()
        expr = unittest.mock.sentinel.expr

        no = xso_query.NotOp(expr)

        ec.eval_bool.return_value = True
        result = no.eval_leaf(ec)
        self.assertFalse(result)

        ec.eval_bool.assert_called_with(expr)

        ec.eval_bool.return_value = False
        result = no.eval_leaf(ec)
        self.assertTrue(result)

        ec.eval_bool.assert_called_with(expr)

    def test_eval_returns_empty_list_on_false(self):
        ec = unittest.mock.sentinel.ec
        expr = unittest.mock.sentinel.expr

        no = xso_query.NotOp(expr)

        with unittest.mock.patch.object(no, "eval_leaf") as eval_leaf:
            eval_leaf.return_value = False

            self.assertSequenceEqual(
                list(no.eval(ec)),
                []
            )

        eval_leaf.assert_called_with(ec)

    def test_eval_returns_list_with_single_True_on_true(self):
        ec = unittest.mock.sentinel.ec
        expr = unittest.mock.sentinel.expr

        no = xso_query.NotOp(expr)

        with unittest.mock.patch.object(no, "eval_leaf") as eval_leaf:
            eval_leaf.return_value = True

            self.assertSequenceEqual(
                list(no.eval(ec)),
                [True]
            )

        eval_leaf.assert_called_with(ec)


class Testnot_(unittest.TestCase):
    def test_uses_as_expr(self):
        expr = unittest.mock.sentinel.expr

        with contextlib.ExitStack() as stack:
            as_expr = stack.enter_context(
                unittest.mock.patch("aioxmpp.xso.query.as_expr")
            )

            NotOp = stack.enter_context(
                unittest.mock.patch("aioxmpp.xso.query.NotOp")
            )

            result = xso_query.not_(expr)

        as_expr.assert_called_with(expr)

        NotOp.assert_called_with(as_expr())

        self.assertEqual(result, NotOp())


class TestConstant(unittest.TestCase):
    def test_is_Expr(self):
        self.assertTrue(issubclass(
            xso_query.Constant,
            xso_query.Expr,
        ))

    def test_eval_returns_value(self):
        ec = unittest.mock.sentinel.ec

        c = xso_query.Constant(unittest.mock.sentinel.value)
        self.assertEqual(
            c.eval(ec),
            [unittest.mock.sentinel.value],
        )


class TestPreExpr(unittest.TestCase):
    def setUp(self):
        class DummyPreExpr(xso_query.PreExpr):
            def __init__(self, *args, mock=None, **kwargs):
                super().__init__(*args, **kwargs)
                self.__mock = mock or unittest.mock.Mock()

            def xq_instantiate(self, ec):
                return self.__mock.xq_instantiate(ec)

        self.DummyPreExpr = DummyPreExpr

    def tearDown(self):
        del self.DummyPreExpr

    def test_has_SoftExprMixin(self):
        self.assertTrue(issubclass(
            xso_query.PreExpr,
            xso_query._SoftExprMixin,
        ))


class TestClass(unittest.TestCase):
    def test_is_PreExpr(self):
        self.assertTrue(issubclass(
            xso_query.Class,
            xso_query.PreExpr,
        ))

    def test_init(self):
        xso_query.Class()

    def test_forwards_all_arguments(self):
        m = unittest.mock.Mock()

        class Foo:
            def __init__(self, *args, **kwargs):
                super().__init__()
                m(args, kwargs)

        class Bar(xso_query.Class, Foo):
            pass

        Bar("a", 1, bar="2")

        m.assert_called_with(("a", 1), {"bar": "2"})

    def test_xq_instantiate(self):
        expr = unittest.mock.sentinel.expr
        cls = xso_query.Class()

        with contextlib.ExitStack() as stack:
            GetInstances = stack.enter_context(
                unittest.mock.patch(
                    "aioxmpp.xso.query.GetInstances"
                )
            )

            result = cls.xq_instantiate(expr)

        GetInstances.assert_called_with(
            expr,
            cls,
        )

        self.assertEqual(
            result,
            GetInstances()
        )

    def test_xq_instantiate_without_expr(self):
        cls = xso_query.Class()

        with contextlib.ExitStack() as stack:
            ContextInstance = stack.enter_context(
                unittest.mock.patch(
                    "aioxmpp.xso.query.ContextInstance"
                )
            )

            result = cls.xq_instantiate()

        ContextInstance.assert_called_with(
            cls,
        )

        self.assertEqual(
            result,
            ContextInstance()
        )


class TestBoundDescriptor(unittest.TestCase):
    def test_is_PreExpr(self):
        self.assertTrue(issubclass(
            xso_query.BoundDescriptor,
            xso_query.PreExpr,
        ))

    def test_has_ExprMixin(self):
        self.assertTrue(issubclass(
            xso_query.BoundDescriptor,
            xso_query._ExprMixin,
        ))

    def test_init(self):
        bd = xso_query.BoundDescriptor(
            unittest.mock.sentinel.class_,
            unittest.mock.sentinel.descriptor,
            unittest.mock.sentinel.expr_class,
        )

        self.assertEqual(
            bd.xq_xso_class,
            unittest.mock.sentinel.class_,
        )

        self.assertEqual(
            bd.xq_descriptor,
            unittest.mock.sentinel.descriptor,
        )
        self.assertEqual(
            bd.xq_expr_class,
            unittest.mock.sentinel.expr_class,
        )

    def test_xq_instantiate(self):
        expr = unittest.mock.sentinel.expr
        class_ = unittest.mock.Mock()
        expr_class = unittest.mock.Mock()

        bd = xso_query.BoundDescriptor(
            class_,
            unittest.mock.sentinel.descriptor,
            expr_class,
        )

        result = bd.xq_instantiate(expr)

        class_.xq_instantiate.assert_called_with(expr)

        expr_class.assert_called_with(
            class_.xq_instantiate(),
            unittest.mock.sentinel.descriptor,
        )

        self.assertEqual(
            result,
            expr_class()
        )

    def test_xq_instantiate_with_extra_args(self):
        expr = unittest.mock.sentinel.expr
        class_ = unittest.mock.Mock()
        expr_class = unittest.mock.Mock()

        bd = xso_query.BoundDescriptor(
            class_,
            unittest.mock.sentinel.descriptor,
            expr_class,
            expr_kwargs={"mapping_type": unittest.mock.sentinel.mt}
        )

        result = bd.xq_instantiate(expr)

        class_.xq_instantiate.assert_called_with(expr)

        expr_class.assert_called_with(
            class_.xq_instantiate(),
            unittest.mock.sentinel.descriptor,
            mapping_type=unittest.mock.sentinel.mt
        )

        self.assertEqual(
            result,
            expr_class()
        )

    def test_bound_descriptor_proxies_to_xq_descriptor(self):
        descriptor = unittest.mock.Mock()

        bd = xso_query.BoundDescriptor(
            unittest.mock.sentinel.class_,
            descriptor,
            unittest.mock.sentinel.expr_class,
        )

        self.assertEqual(
            bd.foo,
            descriptor.foo
        )

        descriptor.foo = "fnord"

        self.assertEqual(
            bd.foo,
            descriptor.foo
        )

        del descriptor.foo

        with self.assertRaises(AttributeError):
            bd.foo


class Testas_expr(unittest.TestCase):
    def test_arbitrary_type(self):
        vs = [
            unittest.mock.sentinel.thing,
            1,
            2,
            "foo",
            1.2,
        ]

        for v in vs:
            with unittest.mock.patch("aioxmpp.xso.query.Constant") as C:
                result = xso_query.as_expr(v)

            C.assert_called_with(v)
            self.assertEqual(
                result,
                C()
            )

    def test_PreExpr(self):
        pe = unittest.mock.Mock(spec=xso_query.PreExpr)
        result = xso_query.as_expr(pe)

        pe.xq_instantiate.assert_called_with(None)

        self.assertEqual(
            result,
            pe.xq_instantiate(),
        )

    def test_PreExpr_with_context(self):
        expr = unittest.mock.sentinel.expr
        pe = unittest.mock.Mock(spec=xso_query.PreExpr)
        result = xso_query.as_expr(pe, expr)

        pe.xq_instantiate.assert_called_with(expr)

        self.assertEqual(
            result,
            pe.xq_instantiate(),
        )

    def test_Expr(self):
        e = unittest.mock.Mock(spec=xso_query.Expr)
        result = xso_query.as_expr(e)

        self.assertIs(result, e)

    def test_Expr_with_Expr_in_expr_attribute(self):
        as_expr = unittest.mock.Mock()
        as_expr.side_effect = xso_query.as_expr

        e_nested = unittest.mock.sentinel.e_nested

        e = unittest.mock.Mock(spec=xso_query.Expr)
        e.expr = e_nested
        self.assertTrue(hasattr(e, "expr"))

        with unittest.mock.patch(
                "aioxmpp.xso.query.as_expr",
                new=as_expr):
            result = xso_query.as_expr(
                e,
                lhs=unittest.mock.sentinel.lhs
            )

        self.assertIs(result, e)

        self.assertSequenceEqual(
            as_expr.mock_calls,
            [
                unittest.mock.call(e,
                                   lhs=unittest.mock.sentinel.lhs),
                unittest.mock.call(e_nested,
                                   lhs=unittest.mock.sentinel.lhs),
            ]
        )


class Test__integration__(unittest.TestCase):
    def test_as_expr_recursively_resolves_lhses(self):
        new_lhs = unittest.mock.sentinel.new_lhs

        cls = xso_query.Class()
        descriptor = xso_query.BoundDescriptor(
            cls,
            unittest.mock.sentinel.descriptor,
            xso_query.GetDescriptor,
        )

        result = xso_query.as_expr(descriptor, lhs=new_lhs)

        self.assertIsInstance(
            result,
            xso_query.GetDescriptor,
        )

        self.assertEqual(
            result.descriptor,
            unittest.mock.sentinel.descriptor,
        )

        self.assertIsInstance(
            result.expr,
            xso_query.GetInstances,
        )

        self.assertEqual(
            result.expr.class_,
            cls,
        )

        self.assertEqual(
            result.expr.expr,
            new_lhs,
        )

    def test_xsopath(self):
        xso = RootXSO()
        xso.attr = "root"

        f = FooXSO()
        f.attr = "foo1"
        xso.children.append(f)

        f = BazXSO()
        f.attr2 = "baz1"
        xso.children.append(f)

        f = FooXSO()
        f.attr = "foo2"
        xso.children.append(f)

        f = FooXSO()
        f.attr = "bar1/foo"
        b = BarXSO()
        b.child = f
        xso.children.append(b)

        ec = xso_query.EvaluationContext()
        ec.set_toplevel_object(xso)

        query = xso_query.as_expr(
            RootXSO.children / BarXSO.child / FooXSO.attr
        )

        self.assertSequenceEqual(
            list(query.eval(ec)),
            ["bar1/foo"]
        )

        query = xso_query.as_expr(
            RootXSO.children / FooXSO.attr
        )

        self.assertSequenceEqual(
            list(query.eval(ec)),
            ["foo1", "foo2"]
        )

        query = xso_query.as_expr(
            RootXSO.children / FooXSO
        )

        self.assertSequenceEqual(
            list(query.eval(ec)),
            list(filter(lambda x: isinstance(x, FooXSO), xso.children))
        )

        query = xso_query.as_expr(
            RootXSO.children / FooXSO[::2]
        )

        self.assertSequenceEqual(
            list(query.eval(ec)),
            [
                xso.children[0],
                xso.children[2],
            ]
        )

        query = xso_query.as_expr(
            RootXSO.children / FooXSO[xso_query.where(FooXSO.attr)][:2]
        )

        self.assertSequenceEqual(
            list(query.eval(ec)),
            [
                xso.children[0],
                xso.children[2],
            ]
        )

        query = xso_query.as_expr(
            RootXSO.children / FooXSO[:2][xso_query.where(FooXSO.attr)]
        )

        self.assertSequenceEqual(
            list(query.eval(ec)),
            [
                xso.children[0],
            ]
        )

        query = xso_query.as_expr(
            RootXSO.children / FooXSO[xso_query.where(FooXSO.attr)][:2] /
            FooXSO.attr
        )

        self.assertSequenceEqual(
            list(query.eval(ec)),
            [
                "foo1",
                "foo2",
            ]
        )

        query = xso_query.as_expr(
            RootXSO.children / FooXSO[:2][xso_query.where(FooXSO.attr)] /
            FooXSO.attr
        )

        self.assertSequenceEqual(
            list(query.eval(ec)),
            [
                "foo1",
            ]
        )

        query = xso_query.as_expr(
            RootXSO.children / FooXSO[:2][
                xso_query.where(xso_query.not_(FooXSO.attr))
            ]
        )

        self.assertSequenceEqual(
            list(query.eval(ec)),
            [
                xso.children[1]
            ]
        )

        query = xso_query.as_expr(
            RootXSO.children / FooXSO[xso_query.where(FooXSO.attr == "foo1")] /
            FooXSO.attr
        )

        self.assertSequenceEqual(
            list(query.eval(ec)),
            [
                "foo1",
            ]
        )

# foo



#!/usr/bin/python3
########################################################################
# File name: form_type_to_code.py
# This file is part of: aioxmpp
#
# LICENSE
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or (at
# your option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
########################################################################
import argparse
import io
import sys

import aioxmpp.xml
import aioxmpp.xso
import aioxmpp.forms
import aioxmpp.forms.xso as forms_xso

from aioxmpp.utils import namespaces


class FormType(aioxmpp.xso.XSO):
    TAG = namespaces.xep0004_data, "form_type"

    fields = aioxmpp.xso.ChildList([forms_xso.Field])

    name = aioxmpp.xso.ChildText(
        (namespaces.xep0004_data, "name"),
    )


CLASS_MAP = {
    forms_xso.FieldType.LIST_SINGLE: aioxmpp.forms.ListSingle,
    forms_xso.FieldType.LIST_MULTI: aioxmpp.forms.ListMulti,

    forms_xso.FieldType.TEXT_SINGLE: aioxmpp.forms.TextSingle,
    forms_xso.FieldType.TEXT_MULTI: aioxmpp.forms.TextMulti,
    forms_xso.FieldType.TEXT_PRIVATE: aioxmpp.forms.TextPrivate,

    forms_xso.FieldType.JID_SINGLE: aioxmpp.forms.JIDSingle,
    forms_xso.FieldType.JID_MULTI: aioxmpp.forms.JIDMulti,

    forms_xso.FieldType.BOOLEAN: aioxmpp.forms.Boolean,
}


def sanitize_field_name(var):
    lhs, sep, rhs = var.rpartition("#")
    if rhs:
        var = rhs
    else:
        var = lhs

    return var.translate(str.maketrans(
        "-:/ ",
        "____",
    ))


def get_field_fqcn(class_):
    return ".".join(["aioxmpp.forms", class_.__qualname__])


def load_definition(infile):
    indata = infile.read()

    # this is nasty
    indata = indata.replace(
        b"<form_type>",
        "<form_type xmlns='{}'>".format(
            namespaces.xep0004_data
        ).encode("utf-8"),
        1
    )

    defn = aioxmpp.xml.read_single_xso(
        io.BytesIO(indata),
        FormType
    )

    return defn


def generate_class(defn, outfile, strip_prefix):
    print("class Form(aioxmpp.forms.Form):", file=outfile)
    print("    FORM_TYPE = {!r}".format(defn.name), file=outfile)
    print(file=outfile)

    for field in defn.fields:
        field_name = sanitize_field_name(field.var)
        if field_name.startswith(strip_prefix):
            field_name = field_name[len(strip_prefix):]
        fqcn = get_field_fqcn(CLASS_MAP[field.type_])

        print(
            "    {name} = {type_}(".format(
                name=field_name,
                type_=fqcn,
            ),
            file=outfile,
        )
        print(
            "        var={var!r},".format(
                var=field.var,
            ),
            file=outfile,
        )

        if field.label:
            print(
                "        label={label!r}".format(
                    label=field.label,
                ),
                file=outfile,
            )

        if field.desc:
            print(
                "        desc={label!r}".format(
                    desc=field.desc,
                ),
                file=outfile,
            )

        print("    )", file=outfile)
        print(file=outfile)


def main():
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "-s", "--strip-prefix",
        default="",
    )
    parser.add_argument(
        "infile",
        type=argparse.FileType("rb"),
    )

    args = parser.parse_args()

    with args.infile as f:
        defn = load_definition(f)

    generate_class(
        defn,
        sys.stdout,
        args.strip_prefix,
    )


if __name__ == "__main__":
    main()



########################################################################
# File name: xso.py
# This file is part of: aioxmpp
#
# LICENSE
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this program.  If not, see
# <http://www.gnu.org/licenses/>.
#
########################################################################
import aioxmpp
import aioxmpp.xso as xso

from aioxmpp.utils import namespaces

namespaces.xep0077_in_band = "jabber:iq:register"

"""
XSO Definitions
===============

.. autoclass:: Query
"""


@aioxmpp.IQ.as_payload_class
class Query(xso.XSO):
    """
    :xep:`077` In-Band Registraion query :class:`~aioxmpp.xso.XSO`.

    .. attribute:: username

    .. attribute:: nick

    .. attribute:: password

    .. attribute:: name

    .. attribute:: first

    .. attribute:: last

    .. attribute:: email

    .. attribute:: address

    .. attribute:: city

    .. attribute:: state

    .. attribute:: zip

    .. attribute:: phone

    .. attribute:: url

    .. attribute:: date

    .. attribute:: misc

    .. attribute:: text

    .. attribute:: key

    .. attribute:: registered

    .. attribute:: remove
    """

    TAG = (namespaces.xep0077_in_band, "query")
    username = xso.ChildText(
        (namespaces.xep0077_in_band, "username"),
        default=None,
    )

    instructions = xso.ChildText(
        (namespaces.xep0077_in_band, "instructions"),
        default=None,
    )

    nick = xso.ChildText(
        (namespaces.xep0077_in_band, "nick"),
        default=None,
    )

    password = xso.ChildText(
        (namespaces.xep0077_in_band, "password"),
        default=None,
    )

    name = xso.ChildText(
        (namespaces.xep0077_in_band, "name"),
        default=None,
    )

    first = xso.ChildText(
        (namespaces.xep0077_in_band, "first"),
        default=None,
    )

    last = xso.ChildText(
        (namespaces.xep0077_in_band, "last"),
        default=None,
    )

    email = xso.ChildText(
        (namespaces.xep0077_in_band, "email"),
        default=None,
    )

    address = xso.ChildText(
        (namespaces.xep0077_in_band, "address"),
        default=None,
    )

    city = xso.ChildText(
        (namespaces.xep0077_in_band, "city"),
        default=None,
    )

    state = xso.ChildText(
        (namespaces.xep0077_in_band, "state"),
        default=None,
    )

    zip = xso.ChildText(
        (namespaces.xep0077_in_band, "zip"),
        default=None,
    )

    phone = xso.ChildText(
        (namespaces.xep0077_in_band, "phone"),
        default=None,
    )

    url = xso.ChildText(
        (namespaces.xep0077_in_band, "url"),
        default=None,
    )

    date = xso.ChildText(
        (namespaces.xep0077_in_band, "date"),
        default=None,
    )

    misc = xso.ChildText(
        (namespaces.xep0077_in_band, "misc"),
        default=None,
    )

    text = xso.ChildText(
        (namespaces.xep0077_in_band, "text"),
        default=None,
    )

    key = xso.ChildText(
        (namespaces.xep0077_in_band, "key"),
        default=None,
    )

    registered = xso.ChildFlag(
        (namespaces.xep0077_in_band, "registered")
    )

    remove = xso.ChildFlag(
        (namespaces.xep0077_in_band, "remove")
    )


    def __init__(self, username=None, password=None, aux_fields=None):
        """
        Get an xso.Query object with the info provided in he parameters.

        :param username: Username of the query
        :type username: :class:`str`
        :param password: Password of the query.
        :type password: :class:`str`
        :param aux_fields: Auxiliary fields in case additional info is needed.
        :type aux_fields: :class:`dict`
        :return: :class:`xso.Query`
        """
        self.username = username
        self.password = password

        if aux_fields is not None:
            for key, value in aux_fields.items():
                setattr(self, key, value)



########################################################################
# File name: test_service.py
# This file is part of: aioxmpp
#
# LICENSE
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this program.  If not, see
# <http://www.gnu.org/licenses/>.
#
########################################################################
import contextlib
import random
import unittest
import unittest.mock

from datetime import timedelta

import aioxmpp
import aioxmpp.service

import aioxmpp.adhoc.service as adhoc_service
import aioxmpp.adhoc.xso as adhoc_xso
import aioxmpp.disco

from aioxmpp.utils import namespaces

from aioxmpp.testutils import (
    make_connected_client,
    CoroutineMock,
    run_coroutine,
)


TEST_PEER_JID = aioxmpp.JID.fromstr("foo@bar.baz/fnord")
TEST_LOCAL_JID = aioxmpp.JID.fromstr("bar@bar.baz/fnord")


class TestAdHocClient(unittest.TestCase):
    def setUp(self):
        self.cc = make_connected_client()
        self.disco_service = unittest.mock.Mock()
        self.disco_service.query_info = CoroutineMock()
        self.disco_service.query_info.side_effect = AssertionError()
        self.disco_service.query_items = CoroutineMock()
        self.disco_service.query_items.side_effect = AssertionError()
        self.c = adhoc_service.AdHocClient(
            self.cc,
            dependencies={
                aioxmpp.disco.DiscoClient: self.disco_service,
            }
        )

    def tearDown(self):
        del self.c, self.cc

    def test_is_service(self):
        self.assertTrue(issubclass(
            adhoc_service.AdHocClient,
            aioxmpp.service.Service,
        ))

    def test_depends_on_disco(self):
        self.assertLess(
            aioxmpp.disco.DiscoClient,
            adhoc_service.AdHocClient,
        )

    def test_detect_support_using_disco(self):
        response = aioxmpp.disco.xso.InfoQuery(
            features={
                "http://jabber.org/protocol/commands",
            }
        )
        self.disco_service.query_info.side_effect = None
        self.disco_service.query_info.return_value = response

        self.assertTrue(
            run_coroutine(self.c.supports_commands(TEST_PEER_JID)),
        )

        self.disco_service.query_info.assert_called_with(
            TEST_PEER_JID,
        )

    def test_detect_absence_of_support_using_disco(self):
        response = aioxmpp.disco.xso.InfoQuery(
            features=set()
        )
        self.disco_service.query_info.side_effect = None
        self.disco_service.query_info.return_value = response

        self.assertFalse(
            run_coroutine(self.c.supports_commands(TEST_PEER_JID)),
        )

        self.disco_service.query_info.assert_called_with(
            TEST_PEER_JID,
        )

    def test_enumerate_commands_uses_disco(self):
        items = [
            aioxmpp.disco.xso.Item(
                TEST_PEER_JID,
                node="cmd{}".format(i),
            )
            for i in range(3)
        ]

        response = aioxmpp.disco.xso.ItemsQuery(
            items=list(items)
        )
        self.disco_service.query_items.side_effect = None
        self.disco_service.query_items.return_value = response

        result = run_coroutine(
            self.c.get_commands(TEST_PEER_JID),
        )

        self.disco_service.query_items.assert_called_with(
            TEST_PEER_JID,
            node="http://jabber.org/protocol/commands",
        )

        self.assertSequenceEqual(
            result,
            items,
        )

    def test_get_command_info_uses_disco(self):
        self.disco_service.query_info.side_effect = None
        self.disco_service.query_info.return_value = \
            unittest.mock.sentinel.result

        result = run_coroutine(
            self.c.get_command_info(TEST_PEER_JID,
                                    unittest.mock.sentinel.node),
        )

        self.disco_service.query_info.assert_called_with(
            TEST_PEER_JID,
            node=unittest.mock.sentinel.node,
        )

        self.assertEqual(
            result,
            unittest.mock.sentinel.result,
        )

    def test_execute(self):
        with unittest.mock.patch(
                "aioxmpp.adhoc.service.ClientSession") as ClientSession:
            ClientSession().start = CoroutineMock()
            ClientSession.reset_mock()
            result = run_coroutine(self.c.execute(
                unittest.mock.sentinel.peer,
                unittest.mock.sentinel.node,
            ))

        ClientSession.assert_called_once_with(
            self.cc.stream,
            unittest.mock.sentinel.peer,
            unittest.mock.sentinel.node,
        )

        ClientSession().start.assert_called_once_with()

        self.assertEqual(result, ClientSession())


class TestCommandNode(unittest.TestCase):
    def test_is_static_node(self):
        self.assertTrue(issubclass(
            adhoc_service.CommandEntry,
            aioxmpp.disco.StaticNode,
        ))

    def test_defaults(self):
        stanza = unittest.mock.Mock()

        cn = adhoc_service.CommandEntry(
            "foo",
            unittest.mock.sentinel.handler,
            features={}
        )

        self.assertDictEqual(
            cn.name,
            {
                None: "foo",
            }
        )

        self.assertIsInstance(
            cn.name,
            aioxmpp.structs.LanguageMap,
        )

        self.assertEqual(
            cn.handler,
            unittest.mock.sentinel.handler
        )

        self.assertIn(
            ("automation", "command-node", None, "foo"),
            list(cn.iter_identities(stanza))
        )

        self.assertCountEqual(
            {
                namespaces.xep0030_info,
                namespaces.xep0050_commands,
            },
            cn.iter_features(unittest.mock.sentinel.stanza)
        )

        self.assertIsNone(
            cn.is_allowed
        )

        self.assertTrue(
            cn.is_allowed_for(unittest.mock.sentinel.jid)
        )

    def test_is_allowed_inhibits_identities_response(self):
        stanza = unittest.mock.Mock()
        is_allowed = unittest.mock.Mock()
        is_allowed.return_value = False

        cn = adhoc_service.CommandEntry(
            "foo",
            unittest.mock.sentinel.handler,
            features={},
            is_allowed=is_allowed
        )

        self.assertSequenceEqual([], list(cn.iter_identities(stanza)))
        is_allowed.assert_called_once_with(
            stanza.from_,
        )

        is_allowed.reset_mock()
        is_allowed.return_value = True

        self.assertIn(
            ("automation", "command-node", None, "foo"),
            list(cn.iter_identities(stanza))
        )

    def test_is_allowed_for_calls_is_allowed_if_defined(self):
        is_allowed = unittest.mock.Mock()

        cn = adhoc_service.CommandEntry(
            "foo",
            unittest.mock.sentinel.handler,
            is_allowed=is_allowed,
        )

        result = cn.is_allowed_for(
            unittest.mock.sentinel.a,
            unittest.mock.sentinel.b,
            x=unittest.mock.sentinel.x,
        )

        is_allowed.assert_called_once_with(
            unittest.mock.sentinel.a,
            unittest.mock.sentinel.b,
            x=unittest.mock.sentinel.x,
        )

        self.assertEqual(
            result,
            is_allowed(),
        )


class TestAdHocServer(unittest.TestCase):
    def setUp(self):
        self.cc = make_connected_client()
        self.cc.local_jid = TEST_LOCAL_JID
        self.disco_service = unittest.mock.Mock()
        self.s = adhoc_service.AdHocServer(
            self.cc,
            dependencies={
                aioxmpp.disco.DiscoServer: self.disco_service,
            }
        )
        self.disco_service.reset_mock()

    def tearDown(self):
        del self.s
        del self.disco_service
        del self.cc

    def test_is_service(self):
        self.assertTrue(issubclass(
            adhoc_service.AdHocServer,
            aioxmpp.service.Service
        ))

    def test_is_disco_node(self):
        self.assertTrue(issubclass(
            adhoc_service.AdHocServer,
            aioxmpp.disco.Node
        ))

    def test_registers_iq_handler(self):
        self.assertTrue(
            aioxmpp.service.is_iq_handler(
                aioxmpp.IQType.SET,
                adhoc_xso.Command,
                adhoc_service.AdHocServer._handle_command,
            )
        )

    def test_registers_as_node(self):
        self.assertIsInstance(
            adhoc_service.AdHocServer.disco_node,
            aioxmpp.disco.mount_as_node,
        )
        self.assertEqual(
            adhoc_service.AdHocServer.disco_node.mountpoint,
            "http://jabber.org/protocol/commands"
        )

    def test_items_empty_by_default(self):
        stanza = unittest.mock.Mock()
        stanza.lang = None

        self.assertSequenceEqual(
            list(self.s.iter_items(stanza)),
            [],
        )

    def test_identity(self):
        self.assertSetEqual(
            {
                ("automation", "command-list", None, None)
            },
            set(self.s.iter_identities(unittest.mock.sentinel.stanza))
        )

    def test_register_stateless_command_makes_it_appear_in_listing(self):
        handler = unittest.mock.Mock()
        base = unittest.mock.Mock()
        base.CommandEntry().name.lookup.return_value = "some name"

        with contextlib.ExitStack() as stack:
            stack.enter_context(
                unittest.mock.patch(
                    "aioxmpp.adhoc.service.CommandEntry",
                    new=base.CommandEntry,
                ),
            )

            self.s.register_stateless_command(
                "node",
                unittest.mock.sentinel.name,
                handler,
            )

        self.assertCountEqual(
            [
                (self.cc.local_jid, "node", "some name"),
            ],
            [
                (item.jid, item.node, item.name)
                for item in self.s.iter_items(base.stanza)
            ]
        )

    def test_listing_respects_is_allowed(self):
        handler = unittest.mock.Mock()
        base = unittest.mock.Mock()
        base.CommandEntry().name.lookup.return_value = "some name"

        with contextlib.ExitStack() as stack:
            stack.enter_context(
                unittest.mock.patch(
                    "aioxmpp.adhoc.service.CommandEntry",
                    new=base.CommandEntry),
            )

            self.s.register_stateless_command(
                "node",
                unittest.mock.sentinel.name,
                handler,
            )

        base.CommandEntry().is_allowed_for.return_value = False

        items = [
            (item.jid, item.node, item.name)
            for item in self.s.iter_items(base.stanza)
        ]

        base.CommandEntry().is_allowed_for.assert_called_once_with(
            base.stanza.from_,
        )

        self.assertCountEqual(
            [
            ],
            items,
        )

    def test_listing_respects_request_language(self):
        handler = unittest.mock.Mock()
        base = unittest.mock.Mock()
        base.stanza.lang = "de"
        base.CommandEntry().name.lookup.return_value = "some name"

        with contextlib.ExitStack() as stack:
            stack.enter_context(
                unittest.mock.patch(
                    "aioxmpp.adhoc.service.CommandEntry",
                    new=base.CommandEntry),
            )

            self.s.register_stateless_command(
                "node",
                unittest.mock.sentinel.name,
                handler,
            )

        base.CommandEntry().is_allowed_for.return_value = True

        items = [
            (item.jid, item.node, item.name)
            for item in self.s.iter_items(base.stanza)
        ]

        base.CommandEntry().name.lookup.assert_called_once_with(
            [
                aioxmpp.structs.LanguageRange.fromstr("de"),
                aioxmpp.structs.LanguageRange.fromstr("en"),
            ]
        )

    def test_register_stateless_registers_command_at_disco_service(self):
        with contextlib.ExitStack() as stack:
            CommandEntry = stack.enter_context(unittest.mock.patch(
                "aioxmpp.adhoc.service.CommandEntry"
            ))

            self.s.register_stateless_command(
                unittest.mock.sentinel.node,
                unittest.mock.sentinel.name,
                unittest.mock.sentinel.handler,
                features=unittest.mock.sentinel.features,
                is_allowed=unittest.mock.sentinel.is_allowed,
            )

        CommandEntry.assert_called_once_with(
            unittest.mock.sentinel.name,
            unittest.mock.sentinel.handler,
            features=unittest.mock.sentinel.features,
            is_allowed=unittest.mock.sentinel.is_allowed,
        )

        self.disco_service.mount_node.assert_called_once_with(
            unittest.mock.sentinel.node,
            CommandEntry()
        )

        (_, (_, obj), _), = self.disco_service.mount_node.mock_calls

    def test__handle_command_raises_item_not_found_for_unknown_node(self):
        req = aioxmpp.IQ(
            type_=aioxmpp.IQType.SET,
            from_=TEST_PEER_JID,
            to=TEST_LOCAL_JID,
            payload=adhoc_xso.Command(
                "node",
            )
        )

        with self.assertRaises(aioxmpp.errors.XMPPCancelError) as ctx:
            run_coroutine(self.s._handle_command(req))

        self.assertEqual(
            ctx.exception.condition,
            aioxmpp.ErrorCondition.ITEM_NOT_FOUND,
        )

        self.assertRegex(
            ctx.exception.text,
            "no such command: 'node'"
        )

    def test__handle_command_raises_forbidden_for_disallowed_node(self):
        handler = CoroutineMock()
        handler.return_value = unittest.mock.sentinel.result
        is_allowed = unittest.mock.Mock()
        is_allowed.return_value = False

        self.s.register_stateless_command(
            "node",
            "Command name",
            handler,
            is_allowed=is_allowed,
        )

        req = aioxmpp.IQ(
            type_=aioxmpp.IQType.SET,
            from_=TEST_PEER_JID,
            to=TEST_LOCAL_JID,
            payload=adhoc_xso.Command(
                "node",
            )
        )

        with self.assertRaises(aioxmpp.errors.XMPPCancelError) as ctx:
            run_coroutine(self.s._handle_command(req))

        is_allowed.assert_called_once_with(req.from_)

        self.assertEqual(
            ctx.exception.condition,
            aioxmpp.ErrorCondition.FORBIDDEN,
        )

        handler.assert_not_called()

    def test__handle_command_dispatches_to_command(self):
        handler = CoroutineMock()
        handler.return_value = unittest.mock.sentinel.result

        self.s.register_stateless_command(
            "node",
            "Command name",
            handler,
        )

        req = aioxmpp.IQ(
            type_=aioxmpp.IQType.SET,
            from_=TEST_PEER_JID,
            to=TEST_LOCAL_JID,
            payload=adhoc_xso.Command(
                "node",
            )
        )

        result = run_coroutine(self.s._handle_command(req))

        handler.assert_called_once_with(req)

        self.assertEqual(
            result,
            unittest.mock.sentinel.result,
        )


class TestClientSession(unittest.TestCase):
    def setUp(self):
        self.stream = unittest.mock.Mock()
        self.send_iq_and_wait_for_reply = CoroutineMock()
        self.send_iq_and_wait_for_reply.return_value = None
        self.stream.send_iq_and_wait_for_reply = \
            self.send_iq_and_wait_for_reply

        self.peer_jid = TEST_PEER_JID
        self.command_name = "foocmd"
        self.session = adhoc_service.ClientSession(
            self.stream,
            self.peer_jid,
            self.command_name,
        )

    def tearDown(self):
        del self.stream
        del self.send_iq_and_wait_for_reply
        del self.session

    def test_init(self):
        self.assertIsNone(
            self.session.status,
        )

        self.assertIsNone(
            self.session.first_payload,
        )

        self.assertIsNone(
            self.session.response,
        )

        self.assertSetEqual(
            self.session.allowed_actions,
            {adhoc_xso.ActionType.EXECUTE,
             adhoc_xso.ActionType.CANCEL}
        )

    def test_start(self):
        response = unittest.mock.Mock()

        self.send_iq_and_wait_for_reply.return_value = response
        self.send_iq_and_wait_for_reply.side_effect = None

        result = run_coroutine(self.session.start())

        self.assertEqual(
            len(self.send_iq_and_wait_for_reply.mock_calls),
            1,
        )

        _, (iq,), _ = self.send_iq_and_wait_for_reply.mock_calls[0]

        self.assertIsInstance(
            iq,
            aioxmpp.IQ,
        )
        self.assertEqual(
            iq.type_,
            aioxmpp.IQType.SET,
        )
        self.assertEqual(
            iq.to,
            self.peer_jid,
        )
        self.assertIsInstance(
            iq.payload,
            adhoc_xso.Command,
        )

        cmd = iq.payload
        self.assertEqual(
            cmd.action,
            adhoc_xso.ActionType.EXECUTE,
        )
        self.assertEqual(
            cmd.node,
            self.command_name,
        )
        self.assertIsNone(cmd.actions)
        self.assertIsNone(cmd.first_payload)
        self.assertSequenceEqual(cmd.notes, [])
        self.assertIsNone(cmd.status)
        self.assertIsNone(cmd.sessionid)

        self.assertEqual(
            result,
            response.first_payload
        )

        self.assertEqual(
            self.session.first_payload,
            response.first_payload
        )

        self.assertEqual(
            self.session.response,
            response
        )

        self.assertEqual(
            self.session.status,
            response.status,
        )

        self.assertEqual(
            self.session.allowed_actions,
            response.actions.allowed_actions,
        )

        self.assertEqual(
            self.session.sessionid,
            response.sessionid,
        )

        # trick
        response.actions = None

        self.assertSetEqual(
            self.session.allowed_actions,
            {adhoc_xso.ActionType.EXECUTE,
             adhoc_xso.ActionType.CANCEL}
        )

    def test_aenter_starts(self):
        with unittest.mock.patch.object(self.session, "start") as start:
            result = run_coroutine(self.session.__aenter__())
            start.assert_called_once_with()
        self.assertEqual(result, self.session)

    def test_aenter_after_start_is_harmless(self):
        response = unittest.mock.Mock()

        self.send_iq_and_wait_for_reply.return_value = response
        self.send_iq_and_wait_for_reply.side_effect = None

        run_coroutine(self.session.start())
        result = run_coroutine(self.session.__aenter__())
        self.assertEqual(result, self.session)

    def test_aexit_closes(self):
        with unittest.mock.patch.object(self.session, "close") as close:
            result = run_coroutine(self.session.__aexit__(
                unittest.mock.sentinel.type_,
                unittest.mock.sentinel.value,
                unittest.mock.sentinel.tb,
            ))
        close.assert_called_once_with()
        self.assertFalse(result)

    def test_reject_start_after_start(self):
        response = unittest.mock.Mock()

        self.send_iq_and_wait_for_reply.return_value = response
        self.send_iq_and_wait_for_reply.side_effect = None

        run_coroutine(self.session.start())

        self.send_iq_and_wait_for_reply.mock_calls.clear()

        with self.assertRaisesRegex(
                RuntimeError,
                r"command execution already started"):
            run_coroutine(self.session.start())

        self.assertSequenceEqual(
            self.send_iq_and_wait_for_reply.mock_calls,
            []
        )

    def test_reject_proceed_before_start(self):
        with self.assertRaisesRegex(
                RuntimeError,
                r"command execution not started yet"):
            run_coroutine(self.session.proceed())

    def test_proceed_uses_execute_and_previous_payload_by_default(self):
        initial_response = unittest.mock.Mock()
        initial_response.payload = [
            unittest.mock.sentinel.payload1,
            unittest.mock.sentinel.payload2,
        ]
        initial_response.sessionid = "foobar"
        initial_response.actions = None
        self.send_iq_and_wait_for_reply.return_value = initial_response
        self.send_iq_and_wait_for_reply.side_effect = None
        run_coroutine(self.session.start())
        self.send_iq_and_wait_for_reply.mock_calls.clear()

        response = unittest.mock.Mock()
        self.send_iq_and_wait_for_reply.return_value = response
        self.send_iq_and_wait_for_reply.side_effect = None

        result = run_coroutine(self.session.proceed())

        _, (iq,), _ = self.send_iq_and_wait_for_reply.mock_calls[0]

        self.assertIsInstance(
            iq,
            aioxmpp.IQ,
        )
        self.assertEqual(
            iq.type_,
            aioxmpp.IQType.SET,
        )
        self.assertEqual(
            iq.to,
            self.peer_jid,
        )
        self.assertIsInstance(
            iq.payload,
            adhoc_xso.Command,
        )

        cmd = iq.payload
        self.assertEqual(
            cmd.action,
            adhoc_xso.ActionType.EXECUTE,
        )
        self.assertEqual(
            cmd.node,
            self.command_name,
        )
        self.assertIsNone(cmd.actions)
        self.assertSequenceEqual(
            cmd.payload,
            initial_response.payload,
        )
        self.assertSequenceEqual(cmd.notes, [])
        self.assertIsNone(cmd.status)
        self.assertEqual(
            cmd.sessionid,
            "foobar",
        )

        self.assertEqual(
            result,
            response.first_payload
        )

        self.assertEqual(
            self.session.first_payload,
            response.first_payload
        )

        self.assertEqual(
            self.session.response,
            response
        )

        self.assertEqual(
            self.session.status,
            response.status,
        )

        self.assertEqual(
            self.session.allowed_actions,
            response.actions.allowed_actions,
        )

    def test_proceed_rejects_disallowed_action(self):
        initial_response = unittest.mock.Mock()
        initial_response.payload = [
            unittest.mock.sentinel.payload1,
            unittest.mock.sentinel.payload2,
        ]
        initial_response.actions.allowed_actions = set()
        self.send_iq_and_wait_for_reply.return_value = initial_response
        self.send_iq_and_wait_for_reply.side_effect = None
        run_coroutine(self.session.start())
        self.send_iq_and_wait_for_reply.mock_calls.clear()

        with self.assertRaisesRegex(
                ValueError,
                r"action .*NEXT not allowed in this stage"):
            run_coroutine(self.session.proceed(
                action=adhoc_xso.ActionType.NEXT
            ))

        self.assertSequenceEqual(
            self.send_iq_and_wait_for_reply.mock_calls,
            []
        )

    def test_proceed_with_custom_action(self):
        initial_response = unittest.mock.Mock()
        initial_response.payload = [
            unittest.mock.sentinel.payload1,
            unittest.mock.sentinel.payload2,
        ]
        initial_response.actions.allowed_actions = {
            adhoc_xso.ActionType.NEXT,
        }
        initial_response.sessionid = "baz"
        self.send_iq_and_wait_for_reply.return_value = initial_response
        self.send_iq_and_wait_for_reply.side_effect = None
        run_coroutine(self.session.start())
        self.send_iq_and_wait_for_reply.mock_calls.clear()

        response = unittest.mock.Mock()
        self.send_iq_and_wait_for_reply.return_value = response
        self.send_iq_and_wait_for_reply.side_effect = None

        run_coroutine(self.session.proceed(
            action=adhoc_xso.ActionType.NEXT,
        ))

        _, (iq,), _ = self.send_iq_and_wait_for_reply.mock_calls[0]
        cmd = iq.payload
        self.assertEqual(
            cmd.action,
            adhoc_xso.ActionType.NEXT,
        )

    def test_proceed_with_custom_payload(self):
        initial_response = unittest.mock.Mock()
        initial_response.payload = [
            unittest.mock.sentinel.payload1,
            unittest.mock.sentinel.payload2,
        ]
        initial_response.actions = None
        initial_response.sessionid = "fnord"
        self.send_iq_and_wait_for_reply.return_value = initial_response
        self.send_iq_and_wait_for_reply.side_effect = None
        run_coroutine(self.session.start())
        self.send_iq_and_wait_for_reply.mock_calls.clear()

        response = unittest.mock.Mock()
        self.send_iq_and_wait_for_reply.return_value = response
        self.send_iq_and_wait_for_reply.side_effect = None

        Command = adhoc_xso.Command
        with unittest.mock.patch(
                "aioxmpp.adhoc.xso.Command") as Command_patched:
            Command_patched.side_effect = Command

            run_coroutine(self.session.proceed(
                payload=unittest.mock.sentinel.payload,
            ))

        Command_patched.assert_called_with(
            self.command_name,
            action=adhoc_xso.ActionType.EXECUTE,
            payload=unittest.mock.sentinel.payload,
            sessionid=initial_response.sessionid,
        )

        _, (iq,), _ = self.send_iq_and_wait_for_reply.mock_calls[0]
        cmd = iq.payload
        self.assertSequenceEqual(
            cmd.payload,
            [unittest.mock.sentinel.payload],
        )

    def test_proceed_calls_close_and_reraises_on_BadSessionID(self):
        initial_response = unittest.mock.Mock()
        initial_response.payload = [
            unittest.mock.sentinel.payload1,
            unittest.mock.sentinel.payload2,
        ]
        initial_response.actions = None
        initial_response.sessionid = "fnord"
        self.send_iq_and_wait_for_reply.return_value = initial_response
        self.send_iq_and_wait_for_reply.side_effect = None
        run_coroutine(self.session.start())
        self.send_iq_and_wait_for_reply.mock_calls.clear()

        exc = aioxmpp.errors.XMPPModifyError(
            aioxmpp.ErrorCondition.BAD_REQUEST,
            text="Bad Session",
            application_defined_condition=adhoc_xso.BadSessionID(),
        )
        self.send_iq_and_wait_for_reply.side_effect = exc

        with unittest.mock.patch.object(
                self.session,
                "close",
                new=CoroutineMock()) as close_:
            with self.assertRaisesRegex(
                    adhoc_service.SessionError,
                    r"Bad Session"):
                run_coroutine(self.session.proceed())

        close_.assert_called_once_with()

    def test_proceed_calls_close_and_reraises_on_SessionExpired(self):
        initial_response = unittest.mock.Mock()
        initial_response.payload = [
            unittest.mock.sentinel.payload1,
            unittest.mock.sentinel.payload2,
        ]
        initial_response.actions = None
        initial_response.sessionid = "fnord"
        self.send_iq_and_wait_for_reply.return_value = initial_response
        self.send_iq_and_wait_for_reply.side_effect = None
        run_coroutine(self.session.start())
        self.send_iq_and_wait_for_reply.mock_calls.clear()

        exc = aioxmpp.errors.XMPPCancelError(
            aioxmpp.ErrorCondition.NOT_ALLOWED,
            text="Session Expired",
            application_defined_condition=adhoc_xso.SessionExpired(),
        )
        self.send_iq_and_wait_for_reply.side_effect = exc

        with unittest.mock.patch.object(
                self.session,
                "close",
                new=CoroutineMock()) as close_:
            with self.assertRaisesRegex(
                    adhoc_service.SessionError,
                    r"Session Expired"):
                run_coroutine(self.session.proceed())

        close_.assert_called_once_with()

    def test_proceed_closes_on_other_cancel_exceptions(self):
        initial_response = unittest.mock.Mock()
        initial_response.payload = [
            unittest.mock.sentinel.payload1,
            unittest.mock.sentinel.payload2,
        ]
        initial_response.actions = None
        initial_response.sessionid = "fnord"
        self.send_iq_and_wait_for_reply.return_value = initial_response
        self.send_iq_and_wait_for_reply.side_effect = None
        run_coroutine(self.session.start())
        self.send_iq_and_wait_for_reply.mock_calls.clear()

        exc = aioxmpp.errors.XMPPCancelError(
            aioxmpp.ErrorCondition.FEATURE_NOT_IMPLEMENTED,
        )
        self.send_iq_and_wait_for_reply.side_effect = exc

        with unittest.mock.patch.object(
                self.session,
                "close",
                new=CoroutineMock()) as close_:
            with self.assertRaises(aioxmpp.errors.XMPPCancelError):
                run_coroutine(self.session.proceed())

        close_.assert_called_once_with()

    def test_proceed_reraises_other_modify_exceptions_without_closing(self):
        initial_response = unittest.mock.Mock()
        initial_response.payload = [
            unittest.mock.sentinel.payload1,
            unittest.mock.sentinel.payload2,
        ]
        initial_response.actions = None
        initial_response.sessionid = "fnord"
        self.send_iq_and_wait_for_reply.return_value = initial_response
        self.send_iq_and_wait_for_reply.side_effect = None
        run_coroutine(self.session.start())
        self.send_iq_and_wait_for_reply.mock_calls.clear()

        exc = aioxmpp.errors.XMPPModifyError(
            aioxmpp.ErrorCondition.BAD_REQUEST,
        )
        self.send_iq_and_wait_for_reply.side_effect = exc

        with unittest.mock.patch.object(
                self.session,
                "close",
                new=CoroutineMock()) as close_:
            with self.assertRaises(aioxmpp.errors.XMPPModifyError):
                run_coroutine(self.session.proceed())

        self.assertFalse(close_.mock_calls)

    def test_allow_close_before_start(self):
        run_coroutine(self.session.close())

        self.assertIsNone(
            self.session.status,
        )

        self.assertIsNone(
            self.session.first_payload,
        )

        self.assertIsNone(
            self.session.response,
        )

        self.assertSetEqual(
            self.session.allowed_actions,
            {adhoc_xso.ActionType.EXECUTE,
             adhoc_xso.ActionType.CANCEL}
        )

    def test_close_after_start_sends_cancel_if_not_completed(self):
        initial_response = unittest.mock.Mock()
        initial_response.payload = [
            unittest.mock.sentinel.payload1,
            unittest.mock.sentinel.payload2,
        ]
        initial_response.actions = None
        initial_response.sessionid = "funk"
        self.send_iq_and_wait_for_reply.return_value = initial_response
        self.send_iq_and_wait_for_reply.side_effect = None
        run_coroutine(self.session.start())
        self.send_iq_and_wait_for_reply.mock_calls.clear()

        response = unittest.mock.Mock()
        self.send_iq_and_wait_for_reply.return_value = response
        self.send_iq_and_wait_for_reply.side_effect = None

        run_coroutine(self.session.close())

        _, (iq,), _ = self.send_iq_and_wait_for_reply.mock_calls[0]

        self.assertIsInstance(
            iq,
            aioxmpp.IQ,
        )
        self.assertEqual(
            iq.type_,
            aioxmpp.IQType.SET,
        )
        self.assertEqual(
            iq.to,
            self.peer_jid,
        )
        self.assertIsInstance(
            iq.payload,
            adhoc_xso.Command,
        )

        cmd = iq.payload
        self.assertEqual(
            cmd.action,
            adhoc_xso.ActionType.CANCEL,
        )
        self.assertEqual(
            cmd.node,
            self.command_name,
        )
        self.assertIsNone(cmd.actions)
        self.assertSequenceEqual(
            cmd.payload,
            [],
        )
        self.assertSequenceEqual(cmd.notes, [])
        self.assertIsNone(cmd.status)
        self.assertEqual(
            cmd.sessionid,
            initial_response.sessionid,
        )

        self.assertIsNone(
            self.session.status,
        )

        self.assertIsNone(
            self.session.first_payload,
        )

        self.assertIsNone(
            self.session.response,
        )

        self.assertSetEqual(
            self.session.allowed_actions,
            {adhoc_xso.ActionType.EXECUTE,
             adhoc_xso.ActionType.CANCEL}
        )

    def test_close_ignores_stanza_errors_in_reply(self):
        initial_response = unittest.mock.Mock()
        initial_response.payload = [
            unittest.mock.sentinel.payload1,
            unittest.mock.sentinel.payload2,
        ]
        initial_response.actions = None
        initial_response.sessionid = "funk"
        self.send_iq_and_wait_for_reply.return_value = initial_response
        self.send_iq_and_wait_for_reply.side_effect = None
        run_coroutine(self.session.start())
        self.send_iq_and_wait_for_reply.mock_calls.clear()

        exc = aioxmpp.errors.StanzaError()
        self.send_iq_and_wait_for_reply.side_effect = exc

        run_coroutine(self.session.close())

        self.assertIsNone(
            self.session.status,
        )

        self.assertIsNone(
            self.session.first_payload,
        )

        self.assertIsNone(
            self.session.response,
        )

        self.assertSetEqual(
            self.session.allowed_actions,
            {adhoc_xso.ActionType.EXECUTE,
             adhoc_xso.ActionType.CANCEL}
        )

    def test_close_does_not_send_cancel_if_completed(self):
        initial_response = unittest.mock.Mock()
        initial_response.payload = [
            unittest.mock.sentinel.payload1,
            unittest.mock.sentinel.payload2,
        ]
        initial_response.actions = None
        initial_response.sessionid = "funk"
        initial_response.status = adhoc_xso.CommandStatus.COMPLETED
        self.send_iq_and_wait_for_reply.return_value = initial_response
        self.send_iq_and_wait_for_reply.side_effect = None
        run_coroutine(self.session.start())
        self.send_iq_and_wait_for_reply.mock_calls.clear()

        run_coroutine(self.session.close())

        self.assertFalse(self.send_iq_and_wait_for_reply.mock_calls)

        self.assertIsNone(
            self.session.status,
        )

        self.assertIsNone(
            self.session.first_payload,
        )

        self.assertIsNone(
            self.session.response,
        )

        self.assertSetEqual(
            self.session.allowed_actions,
            {adhoc_xso.ActionType.EXECUTE,
             adhoc_xso.ActionType.CANCEL}
        )


# class TestServerSession(unittest.TestCase):
#     def setUp(self):
#         self.cc = make_connected_client()
#         self.s = self.cc.stream
#         self.sessionid = "testsessionid"
#         self.ss = adhoc_service.ServerSession(
#             self.s,
#             sessionid=self.sessionid
#         )

#     def tearDown(self):
#         del self.ss
#         del self.sessionid
#         del self.s
#         del self.cc

#     def test_init_uses_system_entropy(self):
#         self.assertIsInstance(
#             adhoc_service._rng,
#             random.SystemRandom,
#         )

#         with unittest.mock.patch("aioxmpp.adhoc.service._rng") as rng:
#             rng.getrandbits.return_value = 1234

#             self.ss = adhoc_service.ServerSession(
#                 self.s,
#                 TEST_PEER_JID,
#             )

#         rng.getrandbits.assert_called_once_with(64)

#         self.assertEqual(
#             self.ss.sessionid,
#             "0gQAAAAAAAA"
#         )

#     def test_init(self):
#         self.assertEqual(self.ss.sessionid, self.sessionid)
#         self.assertEqual(self.ss.timeout, timedelta(seconds=60))

#     def test_reply_raises_if_handle_has_not_been_called(self):
#         with self.assertRaises(RuntimeError):
#             pass



########################################################################
# File name: test_xso.py
# This file is part of: aioxmpp
#
# LICENSE
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this program.  If not, see
# <http://www.gnu.org/licenses/>.
#
########################################################################
import unittest

import aioxmpp.hashes
import aioxmpp.entitycaps.xso as entitycaps_xso
import aioxmpp.stanza as stanza
import aioxmpp.xso as xso

from aioxmpp.utils import namespaces


class TestNamespaces(unittest.TestCase):
    def test_caps(self):
        self.assertEqual(
            namespaces.xep0115_caps,
            "http://jabber.org/protocol/caps"
        )

    def test_ecaps2(self):
        self.assertEqual(
            namespaces.xep0390_caps,
            "urn:xmpp:caps"
        )


class TestCaps115(unittest.TestCase):
    def test_is_xso(self):
        self.assertTrue(issubclass(
            entitycaps_xso.Caps115,
            xso.XSO
        ))

    def test_tag(self):
        self.assertEqual(
            (namespaces.xep0115_caps, "c"),
            entitycaps_xso.Caps115.TAG
        )

    def test_node_attr(self):
        self.assertIsInstance(
            entitycaps_xso.Caps115.node,
            xso.Attr
        )
        self.assertEqual(
            (None, "node"),
            entitycaps_xso.Caps115.node.tag
        )
        self.assertIs(
            entitycaps_xso.Caps115.node.default,
            xso.NO_DEFAULT
        )

    def test_hash_attr(self):
        self.assertIsInstance(
            entitycaps_xso.Caps115.hash_,
            xso.Attr
        )
        self.assertEqual(
            (None, "hash"),
            entitycaps_xso.Caps115.hash_.tag
        )
        self.assertIsInstance(
            entitycaps_xso.Caps115.hash_.validator,
            xso.Nmtoken
        )
        self.assertEqual(
            xso.ValidateMode.FROM_CODE,
            entitycaps_xso.Caps115.hash_.validate
        )
        self.assertIs(
            entitycaps_xso.Caps115.hash_.default,
            None
        )

    def test_ver_attr(self):
        self.assertIsInstance(
            entitycaps_xso.Caps115.ver,
            xso.Attr
        )
        self.assertEqual(
            (None, "ver"),
            entitycaps_xso.Caps115.ver.tag
        )
        self.assertIs(
            entitycaps_xso.Caps115.ver.default,
            xso.NO_DEFAULT
        )

    def test_ext_attr(self):
        self.assertIsInstance(
            entitycaps_xso.Caps115.ext,
            xso.Attr
        )
        self.assertEqual(
            (None, "ext"),
            entitycaps_xso.Caps115.ext.tag
        )
        self.assertIs(
            entitycaps_xso.Caps115.ext.default,
            None
        )

    def test_attr_on_Presence(self):
        self.assertIsInstance(
            stanza.Presence.xep0115_caps,
            xso.Child,
        )
        self.assertSetEqual(
            {
                entitycaps_xso.Caps115
            },
            set(stanza.Presence.xep0115_caps._classes)
        )

    def test_init(self):
        caps = entitycaps_xso.Caps115(
            "node",
            "ver",
            "hash",
        )

        self.assertEqual(caps.node, "node")
        self.assertEqual(caps.ver, "ver")
        self.assertEqual(caps.hash_, "hash")

        with self.assertRaisesRegex(TypeError,
                                    "positional argument"):
            entitycaps_xso.Caps115()


class TestCaps390(unittest.TestCase):
    def test_is_xso(self):
        self.assertTrue(issubclass(
            entitycaps_xso.Caps390,
            xso.XSO,
        ))

    def test_is_hashes_parent(self):
        self.assertTrue(issubclass(
            entitycaps_xso.Caps390,
            aioxmpp.hashes.HashesParent,
        ))

    def test_tag(self):
        self.assertEqual(
            entitycaps_xso.Caps390.TAG,
            (namespaces.xep0390_caps, "c")
        )

    def test_attr_on_Presence(self):
        self.assertIsInstance(
            stanza.Presence.xep0390_caps,
            xso.Child,
        )
        self.assertSetEqual(
            {
                entitycaps_xso.Caps390
            },
            set(stanza.Presence.xep0390_caps._classes)
        )



########################################################################
# File name: service.py
# This file is part of: aioxmpp
#
# LICENSE
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this program.  If not, see
# <http://www.gnu.org/licenses/>.
#
########################################################################

import asyncio
import aioxmpp
import logging
from aioxmpp.service import Service
from . import xso


logger = logging.getLogger(__name__)


@asyncio.coroutine
def get_registration_fields(xmlstream,
                            timeout=60,
                            ):
    """
    A query is sent to the server to obtain the fields that need to be
    filled to register with the server.

    :param xmlstream: Specifies the stream connected to the server where
                      the account will be created.
    :type xmlstream: :class:`aioxmpp.protocol.XMLStream`

    :param timeout: Maximum time in seconds to wait for an IQ response, or
                    :data:`None` to disable the timeout.
    :type timeout: :class:`~numbers.Real` or :data:`None`

    :return: :attr:`list`
    """

    iq = aioxmpp.IQ(
        to=aioxmpp.JID.fromstr(xmlstream._to),
        type_=aioxmpp.IQType.GET,
        payload=xso.Query()
    )
    iq.autoset_id()

    reply = yield from aioxmpp.protocol.send_and_wait_for(xmlstream,
                                                          [iq],
                                                          [aioxmpp.IQ],
                                                          timeout=timeout)
    return reply.payload


@asyncio.coroutine
def register(xmlstream,
             query_xso,
             timeout=60,
             ):
    """
    Create a new account on the server.

    :param query_xso: XSO with the information needed for the registration.
    :type query_xso: :class:`xso.Query`

    :param xmlstream: Specifies the stream connected to the server where
                      the account will be created.
    :type xmlstream: :class:`aioxmpp.protocol.XMLStream`

    :param timeout: Maximum time in seconds to wait for an IQ response, or
                    :data:`None` to disable the timeout.
    :type timeout: :class:`~numbers.Real` or :data:`None`
    """
    iq = aioxmpp.IQ(
        to=aioxmpp.JID.fromstr(xmlstream._to),
        type_=aioxmpp.IQType.SET,
        payload=query_xso
    )
    iq.autoset_id()

    yield from aioxmpp.protocol.send_and_wait_for(xmlstream,
                                                  [iq],
                                                  [aioxmpp.IQ],
                                                  timeout=timeout)


def get_used_fields(payload):
    """
    Get a list containing the names of the fields that are used in the
    xso.Query.

    :param payload: Query object o be
    :type payload: :class:`xso.Query`
    :return: :attr:`list`
    """
    return [
        tag
        for tag, descriptor in payload.CHILD_MAP.items()
        if descriptor.__get__(payload, type(payload)) is not None
    ]

class RegistrationService(Service):
    """
    Service implementing XMPP In-Band Registration(:xep:`0077`).

    This 'service' implements the possibility for an entity to register with a
    XMPP server, cancel an existing registration, or change a password.

    .. automethod:: get_client_info

    .. automethod:: change_pass

    .. automethod:: cancel_registration

    .. automethod:: get_registration_fields

    .. automethod:: register

    """

    @asyncio.coroutine
    def get_client_info(self):
        """
        A query is sent to the server to obtain the client's data stored at the
        server.

        :return: :class:`xso.Query`
        """
        iq = aioxmpp.IQ(
            to=self.client.local_jid.bare().replace(localpart=None),
            type_=aioxmpp.IQType.GET,
            payload=xso.Query()
        )

        reply = (yield from self.client.send(iq))
        return reply

    @asyncio.coroutine
    def change_pass(self,
                    new_pass):
        """
        Change the client password for 'new_pass'.

        :param new_pass: New password of the client.
        :type new_pass: :class:`str`

        :param old_pass: Old password of the client.
        :type old_pass: :class:`str`

        """
        iq = aioxmpp.IQ(
            to=self.client.local_jid.bare().replace(localpart=None),
            type_=aioxmpp.IQType.SET,
            payload=xso.Query(self.client.local_jid.localpart, new_pass)
        )

        yield from self.client.send(iq)

    @asyncio.coroutine
    def cancel_registration(self):
        """
        Cancels the currents client's account with the server.

        Even if the cancelation is succesful, this method will raise an
        exception due to he account no longer exists for the server, so the
        client will fail.
        To continue with the execution, this method should be surrounded by a
        try/except statement.
        """
        iq = aioxmpp.IQ(
            to=self.client.local_jid.bare().replace(localpart=None),
            type_=aioxmpp.IQType.SET,
            payload=xso.Query()
        )

        iq.payload.remove = True
        yield from self.client.send(iq)
